import torch
from torch import nn

import os
import math

import torch
from torch.autograd import Function
from torch.utils.cpp_extension import load
import torch.nn.functional as F


module_path = os.path.dirname(__file__)
_upfirdn2d_op = None

def upfirdn2d_op():
    global _upfirdn2d_op
    if _upfirdn2d_op is None:
        _upfirdn2d_op = load(
            'upfirdn2d',
            sources=[
                os.path.join(module_path, 'upfirdn2d.cpp'),
                os.path.join(module_path, 'upfirdn2d_kernel.cu'),
            ])
    return _upfirdn2d_op

def make_kernel(k):
    k = torch.tensor(k, dtype=torch.float32)

    if k.ndim == 1:
        k = k[None, :] * k[:, None]

    k /= k.sum()

    return k

class UpFirDn2dBackward(Function):
    @staticmethod
    def forward(
        ctx, grad_output, kernel, grad_kernel, up, down, pad, g_pad, in_size, out_size
    ):

        up_x, up_y = up
        down_x, down_y = down
        g_pad_x0, g_pad_x1, g_pad_y0, g_pad_y1 = g_pad

        grad_output = grad_output.reshape(-1, out_size[0], out_size[1], 1)

        grad_input = upfirdn2d_op().upfirdn2d(
            grad_output,
            grad_kernel,
            down_x,
            down_y,
            up_x,
            up_y,
            g_pad_x0,
            g_pad_x1,
            g_pad_y0,
            g_pad_y1,
        )
        grad_input = grad_input.view(in_size[0], in_size[1], in_size[2], in_size[3])

        ctx.save_for_backward(kernel)

        pad_x0, pad_x1, pad_y0, pad_y1 = pad

        ctx.up_x = up_x
        ctx.up_y = up_y
        ctx.down_x = down_x
        ctx.down_y = down_y
        ctx.pad_x0 = pad_x0
        ctx.pad_x1 = pad_x1
        ctx.pad_y0 = pad_y0
        ctx.pad_y1 = pad_y1
        ctx.in_size = in_size
        ctx.out_size = out_size

        return grad_input

    @staticmethod
    def backward(ctx, gradgrad_input):
        kernel, = ctx.saved_tensors

        gradgrad_input = gradgrad_input.reshape(-1, ctx.in_size[2], ctx.in_size[3], 1)

        gradgrad_out = upfirdn2d_op().upfirdn2d(
            gradgrad_input,
            kernel,
            ctx.up_x,
            ctx.up_y,
            ctx.down_x,
            ctx.down_y,
            ctx.pad_x0,
            ctx.pad_x1,
            ctx.pad_y0,
            ctx.pad_y1,
        )
        # gradgrad_out = gradgrad_out.view(ctx.in_size[0], ctx.out_size[0], ctx.out_size[1], ctx.in_size[3])
        gradgrad_out = gradgrad_out.view(
            ctx.in_size[0], ctx.in_size[1], ctx.out_size[0], ctx.out_size[1]
        )

        return gradgrad_out, None, None, None, None, None, None, None, None


class UpFirDn2d(Function):
    @staticmethod
    def forward(ctx, input, kernel, up, down, pad):
        up_x, up_y = up
        down_x, down_y = down
        pad_x0, pad_x1, pad_y0, pad_y1 = pad

        kernel_h, kernel_w = kernel.shape
        batch, channel, in_h, in_w = input.shape
        ctx.in_size = input.shape

        input = input.reshape(-1, in_h, in_w, 1)

        ctx.save_for_backward(kernel, torch.flip(kernel, [0, 1]))

        out_h = (in_h * up_y + pad_y0 + pad_y1 - kernel_h) // down_y + 1
        out_w = (in_w * up_x + pad_x0 + pad_x1 - kernel_w) // down_x + 1
        ctx.out_size = (out_h, out_w)

        ctx.up = (up_x, up_y)
        ctx.down = (down_x, down_y)
        ctx.pad = (pad_x0, pad_x1, pad_y0, pad_y1)

        g_pad_x0 = kernel_w - pad_x0 - 1
        g_pad_y0 = kernel_h - pad_y0 - 1
        g_pad_x1 = in_w * up_x - out_w * down_x + pad_x0 - up_x + 1
        g_pad_y1 = in_h * up_y - out_h * down_y + pad_y0 - up_y + 1

        ctx.g_pad = (g_pad_x0, g_pad_x1, g_pad_y0, g_pad_y1)

        out = upfirdn2d_op().upfirdn2d(
            input, kernel, up_x, up_y, down_x, down_y, pad_x0, pad_x1, pad_y0, pad_y1
        )
        # out = out.view(major, out_h, out_w, minor)
        out = out.view(-1, channel, out_h, out_w)

        return out

    @staticmethod
    def backward(ctx, grad_output):
        kernel, grad_kernel = ctx.saved_tensors

        grad_input = UpFirDn2dBackward.apply(
            grad_output,
            kernel,
            grad_kernel,
            ctx.up,
            ctx.down,
            ctx.pad,
            ctx.g_pad,
            ctx.in_size,
            ctx.out_size,
        )

        return grad_input, None, None, None, None


def upfirdn2d(input, kernel, up=1, down=1, pad=(0, 0)):
    out = UpFirDn2d.apply(
        input, kernel, (up, up), (down, down), (pad[0], pad[1], pad[0], pad[1])
    )

    return out


def upfirdn2d_native(
    input, kernel, up_x, up_y, down_x, down_y, pad_x0, pad_x1, pad_y0, pad_y1
):
    _, in_h, in_w, minor = input.shape
    kernel_h, kernel_w = kernel.shape

    out = input.view(-1, in_h, 1, in_w, 1, minor)
    out = F.pad(out, [0, 0, 0, up_x - 1, 0, 0, 0, up_y - 1])
    out = out.view(-1, in_h * up_y, in_w * up_x, minor)

    out = F.pad(
        out, [0, 0, max(pad_x0, 0), max(pad_x1, 0), max(pad_y0, 0), max(pad_y1, 0)]
    )
    out = out[
        :,
        max(-pad_y0, 0) : out.shape[1] - max(-pad_y1, 0),
        max(-pad_x0, 0) : out.shape[2] - max(-pad_x1, 0),
        :,
    ]

    out = out.permute(0, 3, 1, 2)
    out = out.reshape(
        [-1, 1, in_h * up_y + pad_y0 + pad_y1, in_w * up_x + pad_x0 + pad_x1]
    )
    w = torch.flip(kernel, [0, 1]).view(1, 1, kernel_h, kernel_w)
    out = F.conv2d(out, w)
    out = out.reshape(
        -1,
        minor,
        in_h * up_y + pad_y0 + pad_y1 - kernel_h + 1,
        in_w * up_x + pad_x0 + pad_x1 - kernel_w + 1,
    )
    out = out.permute(0, 2, 3, 1)

    return out[:, ::down_y, ::down_x, :]



class EqualLinear(nn.Module):
    def __init__(
        self, in_dim, out_dim, bias=True, bias_init=0, lr_mul=1, activation=None
    ):
        super().__init__()

        self.weight = nn.Parameter(torch.randn(out_dim, in_dim).div_(lr_mul))

        if bias:
            self.bias = nn.Parameter(torch.zeros(out_dim).fill_(bias_init))

        else:
            self.bias = None

        self.activation = activation

        self.scale = (1 / math.sqrt(in_dim)) * lr_mul
        self.lr_mul = lr_mul

    def forward(self, input):
        if self.activation:
            out = F.linear(input, self.weight * self.scale)
            out = fused_leaky_relu(out, self.bias * self.lr_mul)

        else:
            out = F.linear(
                input, self.weight * self.scale, bias=self.bias * self.lr_mul
            )

        return out


class Blur(nn.Module):
    def __init__(self, kernel, pad, upsample_factor=1):
        super().__init__()

        kernel = make_kernel(kernel)

        if upsample_factor > 1:
            kernel = kernel * (upsample_factor ** 2)

        self.register_buffer('kernel', kernel)

        self.pad = pad

    def forward(self, input):
        out = upfirdn2d(input, self.kernel, pad=self.pad)

        return out


class ModulatedConv2d(nn.Module):
    def __init__(
            self,
            in_channel,
            out_channel,
            kernel_size,
            style_dim,
            demodulate=True,
            upsample=True,
            downsample=False,
            lr_mul=1.0,
            blur_kernel=[1, 3, 3, 1],
            ):
        super(ModulatedConv2d, self).__init__()

        self.eps = 1e-8
        self.kernel_size = kernel_size
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.upsample = upsample
        self.downsample = downsample

        if upsample:
            factor = 2
            p = (len(blur_kernel) - factor) - (kernel_size - 1)
            pad0 = (p + 1) // 2 + factor - 1
            pad1 = p // 2 + 1

            self.blur = Blur(blur_kernel, pad=(pad0, pad1), upsample_factor=factor)

        if downsample:
            factor = 2
            p = (len(blur_kernel) - factor) + (kernel_size - 1)
            pad0 = (p + 1) // 2
            pad1 = p // 2

            self.blur = Blur(blur_kernel, pad=(pad0, pad1))

        fan_in = in_channel * kernel_size ** 2
        self.scale = 1 / math.sqrt(fan_in)
        self.padding = kernel_size // 2

        self.mod_weight = nn.Parameter(
            torch.randn(1, out_channel, in_channel, kernel_size, kernel_size)
        )

        self.modulation = EqualLinear(style_dim, in_channel, bias_init=1, lr_mul=lr_mul)

        self.demodulate = demodulate

    def forward(self, input, style):
        batch, in_channel, height, width = input.shape

        style = self.modulation(style).view(batch, 1, in_channel, 1, 1)
        weight = self.scale * self.mod_weight * style

        if self.demodulate:
            demod = torch.rsqrt(weight.pow(2).sum([2, 3, 4]) + 1e-8)
            weight = weight * demod.view(batch, self.out_channel, 1, 1, 1)

        weight = weight.view(
            batch * self.out_channel, in_channel, self.kernel_size, self.kernel_size
        )

        if self.upsample:
            input = input.reshape(1, batch * in_channel, height, width)
            weight = weight.view(
                batch, self.out_channel, in_channel, self.kernel_size, self.kernel_size
            )
            weight = weight.transpose(1, 2).reshape(
                batch * in_channel, self.out_channel, self.kernel_size, self.kernel_size
            )
            out = F.conv_transpose2d(input, weight, padding=0, stride=2, groups=batch)
            _, _, height, width = out.shape
            out = out.view(batch, self.out_channel, height, width)
            out = self.blur(out)

        elif self.downsample:
            input = self.blur(input)
            _, _, height, width = input.shape
            input = input.view(1, batch * in_channel, height, width)
            out = F.conv2d(input, weight, padding=0, stride=2, groups=batch)
            _, _, height, width = out.shape
            out = out.view(batch, self.out_channel, height, width)

        else:
            input = input.reshape(1, batch * in_channel, height, width)
            out = F.conv2d(input, weight, padding=self.padding, groups=batch)
            _, _, height, width = out.shape
            out = out.view(batch, self.out_channel, height, width)

        return out





#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class ForceEquilibriumTrainHook(BaseTrainHook):
  "Forces d_fake close to d_real iff too far apart"
  def __init__(self, gan=None, config=None, trainer=None, name="GpSnMemoryTrainHook", distance=0.001, lam=1.0):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)

    klip = self.gan.configurable_param(distance)
    k_lip = (lam * tf.nn.relu(tf.abs(tf.reduce_mean(self.gan.loss.d_real-self.gan.loss.d_fake))-klip))
    self.gan.add_metric("force_eq", k_lip)
    self.loss = [None, k_lip]

  def losses(self):
      return self.loss

  def distributed_step(self, input_iterator_next):
      return [tf.no_op()]
  def distributed_initial_step(self, input_iterator_next):
      return [tf.no_op()]



from PIL import Image
from hypergan.samplers.base_sampler import BaseSampler
from hypergan.viewer import GlobalViewer
import numpy as np
import torch
import torch.nn as nn
import time

class BatchWalkSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=4, session=None):
        BaseSampler.__init__(self, gan, samples_per_row)
        self.latent1 = self.gan.latent.next()
        self.latent2 = self.gan.latent.next()
        self.velocity = 10/24.0
        direction = self.gan.latent.next()
        self.pos = self.latent1
        self.direction = direction / torch.norm(direction, p=2, dim=1, keepdim=True).expand_as(direction)
        self.hardtanh = nn.Hardtanh()
        self.ones = torch.ones_like(self.direction, device="cuda:0")

    def compatible_with(gan):
        if hasattr(gan, 'latent'):
            return True
        return False

    def _sample(self):
        gan = self.gan

        self.pos = self.direction * self.velocity + self.pos
        mask = torch.gt(self.pos, self.ones)
        mask += torch.lt(self.pos, -self.ones)
        self.direction = self.direction + 2 * self.direction * (-self.ones * mask)

        g = gan.generator.forward(self.pos)
        #    gs.append(g)
        return [
            ('generator', g)
        ]



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class EvolutionTrainer(BaseTrainer):
    def _create(self):
        gan = self.gan
        generator = self.gan.generator
        config = self.config

        d_vars = self.d_vars or gan.discriminator.variables()

        loss = self.loss or gan.loss
        d_loss, g_loss = loss.sample

        self.d_log = -tf.log(tf.abs(d_loss+TINY))


        d_optimizer = self.build_optimizer(config, 'd_', config.d_trainer, self.d_lr, d_vars, d_loss)
        #TODO more than one g_loss
        g_optimizer = [self.build_optimizer(config, 'g_', config.g_trainer, self.g_lr, g.variables(), g_loss) for g, l in zip(generator.children, loss.children_losses)]

        assign_children = []
        for p, o in generator.parent_child_tuples:
            for ov, pv in zip(o.variables(), p.variables()):
                op=tf.assign(ov, pv)
                if config.mutation_percent:
                    op += tf.random_normal(self.gan.ops.shape(pv), mean=0, stddev=0.01) * tf.cast(tf.greater(config.mutation_percent, tf.random_uniform(shape=self.gan.ops.shape(pv), minval=0, maxval=1)), tf.float32)
                assign_children.append(op)
        self.clone_parent = tf.group(*assign_children)


        update_parent=[]
        for p, o in generator.parent_child_tuples:
            c_to_p = []
            for ov, pv in zip(o.variables(), p.variables()):
                op=tf.assign(pv, ov)
                c_to_p.append(op)
            update_parent.append(tf.group(*c_to_p))
        self.update_parent = update_parent
        f_lambda = config.f_lambda or 1

        def _squash(grads):
            return tf.add_n([tf.reshape(gan.ops.squash(g), [1]) for g in grads])
        children_grads = [_squash(tf.gradients(l, d_vars)) for l in loss.children_losses]
        if config.fitness == "g":
            self.measure_g = [-l for l in loss.children_losses]
        else:
            self.measure_g = [-l+f_lambda*(-tf.log(TINY+grad_d - tf.log(TINY+tf.nn.sigmoid(loss.d_loss)) - tf.log(TINY+1-tf.nn.sigmoid(l)))) for l, grad_d in zip(loss.children_losses, children_grads)]
        loss.metrics['measure_g'] = tf.reduce_mean(self.measure_g)
        loss.metrics['g_loss'] = loss.g_loss
        loss.metrics['d_loss'] = loss.d_loss

        self.g_loss = g_loss
        self.d_loss = d_loss
        self.d_optimizer = d_optimizer
        self.g_optimizer = g_optimizer
        self.hist = [0 for i in range(len(self.gan.generator.children))]

        return g_optimizer, d_optimizer

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = self.loss or gan.loss
        metrics = loss.metrics
        generator = gan.generator

        d_loss, g_loss = loss.sample

        #winner = np.random.choice(range(len(gan.generator.children)))
        winners = []
        
        for i in range(len(generator.parents)):
            child_count = generator.config.child_count
            choices = self.measure_g[i*child_count:(i+1)*child_count]
            choice = np.argmax(sess.run(choices))
            winner = i*child_count + choice
            self.hist[winner]+=1
            winners.append(winner)
        sess.run([self.update_parent[winner] for winner in winners])
        for i in range(config.d_update_steps or 1):
            sess.run(self.d_optimizer)

        sess.run(self.clone_parent)
        for i in range(config.g_update_steps or 1):
            sess.run(self.g_optimizer)
        measure_g = sess.run(self.measure_g)

        if self.current_step % 100 == 0:
            hist_output = "  " + "".join(["G"+str(i)+":"+str(v)+" "for i, v in enumerate(self.hist)])
            metric_values = sess.run(self.output_variables(metrics), feed_dict)
            print(str(self.output_string(metrics) % tuple([self.current_step] + metric_values)+hist_output))
            self.hist = [0 for i in range(len(self.gan.generator.children))]




from hypergan.samplers.base_sampler import BaseSampler

class InputSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)

    def _sample(self):
        return {
            'generator': self.gan.inputs.next()[0]
        }




import numpy as np
import torch
import hyperchamber as hc
import inspect
from torch.autograd import Variable

from hypergan.trainers.alternating_trainer import AlternatingTrainer

class AccumulateGradientTrainer(AlternatingTrainer):
    """ G gradients accumulate over many D steps """
    def _create(self):
        self.d_optimizer = self.create_optimizer("d_optimizer")
        self.g_optimizer = self.create_optimizer("g_optimizer")
        self.accumulated_g_grads = None
        self.accumulation_steps = 0
        self.max_grads = None
        self.min_grads = None
        self.max_variance = None
        self.relu = torch.nn.ReLU()

    def calculate_gradients(self):
        accumulate = (self.config.accumulate or 3)
        if self.accumulation_steps == accumulate:
            if self.config.type == 'variance':
                for i, g in enumerate(self.accumulated_g_grads):
                    range_stddev = (self.max_grads[i]-self.min_grads[i])/4.0
                    spread_ratio = (range_stddev / (g+1e-12)).abs()
                    doubt = torch.clamp(self.relu((self.config.allowed_variance or 1.0) - spread_ratio/(self.config.max_spread or 0.2)), max=1.0)
                    if self.config.verbose:
                        print("confidence >>", i, doubt.sum(), "/", np.prod(g.shape), "=", ("%d" % (doubt.sum()/np.prod(g.shape) * 100.0).item())+"%")
                    self.accumulated_g_grads[i] = g * doubt
            g_grads = self.accumulated_g_grads
            d_grads = []
            #print("G_G", sum([g.abs().sum() for g in self.accumulated_g_grads[0]]), len(self.accumulated_g_grads))
            self.accumulated_g_grads = None
            self.accumulation_steps = 0
        else:
            gs = self.g_grads()
            if self.accumulated_g_grads is None:
                self.accumulated_g_grads = [g.clone()/accumulate for g in gs]
                if self.config.type == 'variance':
                    self.max_grads = [g.clone() for g in gs]
                    self.min_grads = [g.clone() for g in gs]
            else:
                for i, g in enumerate(self.accumulated_g_grads):
                    if self.config.type == 'agree':
                        self.accumulated_g_grads[i] = (self.accumulated_g_grads[i] + gs[i].clone()/accumulate) * self.relu(torch.sign(self.accumulated_g_grads[i]*gs[i].clone()))
                    else:
                        if self.config.type == 'variance':
                            self.max_grads[i] = torch.max(self.max_grads[i], gs[i].clone())
                            self.min_grads[i] = torch.min(self.min_grads[i], gs[i].clone())
                        self.accumulated_g_grads[i] += gs[i].clone() / accumulate

            #print("D_G", sum([g.abs().sum() for g in gs]), len(self.accumulated_g_grads))

            d_grads = self.d_grads()
            g_grads = []
            self.accumulation_steps += 1

        return d_grads, g_grads


    def _step(self, feed_dict):
        metrics = self.gan.metrics()

        self.before_step(self.current_step, feed_dict)

        d_grads, g_grads = self.calculate_gradients()
        self.train_d(d_grads)
        self.train_g(g_grads)

        self.after_step(self.current_step, feed_dict)

        if self.current_step % 20 == 0:
            self.print_metrics(self.current_step)





import numpy as np
import torch
import hyperchamber as hc
import inspect
from torch.autograd import Variable

from hypergan.trainers.alternating_trainer import AlternatingTrainer

TINY = 1e-12

class BalancedTrainer(AlternatingTrainer):
    """ G or D depending on which scores better """
    def _create(self):
        self.d_optimizer = self.create_optimizer("d_optimizer")
        self.g_optimizer = self.create_optimizer("g_optimizer")
        self.dcount = 0
        self.gcount = 0
        self.last_d_fake = None

    def calculate_gradients(self):
        d_real, d_fake = self.gan.forward_pass()
        self.add_metric("R", d_real.mean())
        self.add_metric("F", d_fake.mean())

        if (self.config.pretrain_d or 0) > self.gan.steps:
            step_d = True
        elif self.config.d_fake_balance:
            if self.last_d_fake is None or d_fake.mean() > self.last_d_fake:
                step_d = True
            else:
                step_d = False
            self.last_d_fake = d_fake.mean()
        elif d_real.mean() < (d_fake.mean()+(self.config.imbalance or 0.1)):
            step_d = True
        else:
            step_d = False

        if step_d:
            d_grads = self.d_grads()
            g_grads = []

            self.dcount+=1
        else:
            d_grads = []
            g_grads = self.g_grads()
            self.gcount+=1
        self.gan.add_metric("dcount", self.dcount)
        self.gan.add_metric("gcount", self.gcount)

        return d_grads, g_grads


    def _step(self, feed_dict):
        metrics = self.gan.metrics()

        self.before_step(self.current_step, feed_dict)

        d_grads, g_grads = self.calculate_gradients()
        self.train_d(d_grads)
        self.train_g(g_grads)

        self.after_step(self.current_step, feed_dict)

        if self.current_step % 20 == 0:
            self.print_metrics(self.current_step)





import numpy as np
import torch
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class AlternatingTrainer(BaseTrainer):
    """ Steps G and D alternating """
    def _create(self):
        self.d_optimizer = self.create_optimizer("d_optimizer")
        self.g_optimizer = self.create_optimizer("g_optimizer")

    def required(self):
        return "d_optimizer g_optimizer".split()

    def g_grads(self, d_real=None, d_fake=None):
        self.g_optimizer.zero_grad()
        self.setup_gradient_flow(self.gan.g_parameters(), self.gan.d_parameters())

        if d_fake is None:
            _, g_loss = self.gan.forward_loss()#TODO targets=['d']
        else:
            _, g_loss = self.gan.loss.forward(d_real, d_fake)#TODO targets=['d']
        _, g_loss = self.gan.forward_loss()#TODO targets=['d']

        self.gan.add_metric('g_loss', g_loss.mean())
        g_loss += sum([l[0] for l in self.train_hook_losses(None, g_loss) if l[0] is not None])
        g_loss = g_loss.mean()

        return self.grads_for(g_loss, self.gan.g_parameters())

    def d_grads(self, d_real=None, d_fake=None):
        self.d_optimizer.zero_grad()
        self.setup_gradient_flow(self.gan.d_parameters(), self.gan.g_parameters())

        if d_fake is None:
            d_loss, _ = self.gan.forward_loss()#TODO targets=['d']
        else:
            d_loss, _ = self.gan.loss.forward(d_real, d_fake)#TODO targets=['d']
        self.gan.add_metric('d_loss', d_loss.mean())
        d_loss += sum([l[0] for l in self.train_hook_losses(d_loss, None) if l[0] is not None])
        d_loss = d_loss.mean()

        return self.grads_for(d_loss, self.gan.d_parameters())

    def setup_gradient_flow(self, train_params, ignore_params):
        for p in train_params:
            p.requires_grad = True
        for p in ignore_params:
            p.requires_grad = False

    def grads_for(self, loss, train_params):
        if loss == 0:
            return []
        loss.backward(retain_graph=True)
        #return torch_grad(outputs=loss, inputs=train_params, retain_graph=True)
        return [p.grad for p in train_params]

    def train_hook_losses(self, d_loss, g_loss):
        losses = []
        for hook in self.train_hooks:
            losses.append(hook.forward(d_loss, g_loss))#TODO targets=['d']
        return losses

    def calculate_gradients(self, targets=['d','g']):
        g_grads = []
        if 'g' in targets:
            if( self.config.train_g_every is None or
                (self.gan.steps % self.config.train_g_every == 0)):
                g_grads = self.g_grads()
            if (self.config.pretrain_d is not None and
                self.config.pretrain_d > self.gan.steps[0]):
                g_grads = []

        d_grads = []
        if 'd' in targets:
            if ( self.config.train_d_every is None or
               ( self.gan.steps % self.config.train_d_every == 0)):
                d_grads = self.d_grads()

            if (self.config.pretrain_d is not None and
                self.config.pretrain_d > self.gan.steps):
                d_grads = self.d_grads()

        return d_grads, g_grads

    def train_d(self, grads):
        if(len(grads) == 0):
            return

        for hook in self.train_hooks:
            d_grads, _ = hook.gradients(grads, [])
        for p, np in zip(self.gan.d_parameters(), grads):
            p.grad = np

        self.d_optimizer.step()

    def train_g(self, grads):
        if(len(grads) == 0):
            return

        for hook in self.train_hooks:
            _, grads = hook.gradients([], grads)
        for p, np in zip(self.gan.g_parameters(), grads):
            p.grad = np

        self.g_optimizer.step()

    def _step(self, feed_dict):
        metrics = self.gan.metrics()

        self.before_step(self.current_step, feed_dict)

        d_grads, _ = self.calculate_gradients(['d'])
        self.train_d(d_grads)
        _, g_grads = self.calculate_gradients(['g'])
        self.train_g(g_grads)

        self.after_step(self.current_step, feed_dict)

        if self.current_step % 10 == 0:
            self.print_metrics(self.current_step)


    def print_metrics(self, step):
        metrics = self.gan.metrics()
        metric_values = self.output_variables(metrics)
        print(str(self.output_string(metrics) % tuple([step] + metric_values)))




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect
from hypergan.gan_component import ValidationException

class ElasticWeightConsolidationOptimizer(optimizer.Optimizer):
  """ From https://arxiv.org/abs/1612.00796 """
  def __init__(self, learning_rate=0.001, loss=None, p=0.1, gan=None, config=None, use_locking=False, name="ElasticWeightConsolidationOptimizer", optimizer=None, rho=1, beta=1, gamma=1):
    super().__init__(use_locking, name)
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate
    self.loss = loss

    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    d_vars = []
    g_vars = []
    d_grads = []
    g_grads = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
            d_grads += [grad]
        elif var in self.gan.g_vars():
            g_vars += [var]
            g_grads += [grad]
        else:
            raise ValidationException("Couldn't find var in g_vars or d_vars " + var.name)
    all_grads = d_grads + g_grads
    var_list = d_vars + g_vars

    with ops.init_scope():
        initial_f1 = [tf.constant(self.config.initial_constraint or 0.0, shape=self.gan.ops.shape(v)) for v in var_list]
        f1 = [self._get_or_make_slot(v, f, "f", self._name) for f,v in zip(initial_f1, var_list)]
        v1 = [self._get_or_make_slot(v, v, "v1", self._name) for v in var_list]
        if self.config.include_slots:
            for name in self.optimizer.get_slot_names():
                for var in self.optimizer.variables():
                    self._zeros_slot(var, "pm", "pm")
    self._prepare()

    f1 = [self.get_slot(v, "f") for v in var_list]
    v1 = [self.get_slot(v, "v1") for v in var_list]
    slots_list = []
    slots_vars = []
    if self.config.include_slots:
        for name in self.optimizer.get_slot_names():
            for var in self.optimizer.variables():
                slots_vars += [var]
                slots_list.append(self._zeros_slot(var, "pm", "pm"))

    current_vars = var_list + slots_vars
    tmp_vars = v1 + slots_list

    diff = [tf.square(v-t) for v,t in zip(current_vars, tmp_vars)]

    f_accum = []
    f_decay = self.gan.configurable_param(self.config.f_decay or 0.95)
    gradient_scale = self.gan.configurable_param(self.config.gradient_scale or 1.0)
    for v, f, g in zip(var_list, f1, all_grads):
        opts = self.gan.layer_options(v)
        if opts is not None and "ewc_f_decay" in opts:
          f_decay = self.gan.configurable_param(opts["ewc_f_decay"])
          print("Setting f_decay to ", f_decay, " for ", v)

        if opts is not None and "ewc_gradient_scale" in opts:
          gradient_scale = self.gan.configurable_param(opts["ewc_gradient_scale"])
          print("Setting gradient_scale to ", gradient_scale, " for ", v)
        f_accum += [f_decay * f + gradient_scale * tf.square(g)]
    #f_accum = [tf.where(tf.is_nan(_f), tf.zeros_like(_f), _f) for _f in f_accum]
    #f_accum = [tf.where(tf.is_inf(_f), tf.zeros_like(_f), _f) for _f in f_accum]
    self.gan.add_metric('f1',tf.reduce_sum([tf.reduce_sum(f) for f in f_accum]))

    reg = [tf.multiply(f, d) for f,d in zip(f1, diff)]
    #reg = [tf.where(tf.is_nan(_f), tf.zeros_like(_f), _f) for _f in reg]
    ewc_loss = self.gan.configurable_param(self.config.lam or 17.5)/2.0 * tf.reduce_sum(tf.add_n([tf.reduce_sum(r) for r in reg]))
    self.gan.add_metric('ewc',ewc_loss)

    save_weights = tf.group(*[tf.assign(w, v) for w,v in zip(tmp_vars, current_vars)]) # store variables

    if isinstance(self.loss, list):
        if self.config.add_ewc_loss_gradients:
            newloss = [ewc_loss, ewc_loss]
        else:
            newloss = [self.loss[0]+ewc_loss, self.loss[1]+ewc_loss]

        new_grads = tf.gradients(newloss[0], d_vars) + tf.gradients(newloss[1], g_vars)
        self.optimizer.loss = [ewc_loss+self.loss[0], ewc_loss+self.loss[1]]
    else:
        if self.config.add_ewc_loss_gradients:
            newloss = ewc_loss
        else:
            newloss = self.loss+ewc_loss

        new_grads = tf.gradients(newloss, current_vars)
        self.optimizer.loss =ewc_loss+self.loss

    if self.config.add_ewc_loss_gradients:
        new_grads = [_g+_ng for _g,_ng in zip(all_grads, new_grads)]

    for g, oldg, v in zip(new_grads, all_grads, current_vars):
        if(self.gan.ops.shape(g) != self.gan.ops.shape(oldg)):
            print("[ERROR] Shape change on gradients for", v, g, "old g", oldg)
            raise "Gradient change error"
    step = self.optimizer.apply_gradients(list(zip(new_grads, current_vars)).copy(), global_step=global_step, name=name)

    store_f = tf.group(*[tf.assign(w, v) for w,v in zip(f1, f_accum)])
    with tf.get_default_graph().control_dependencies([store_f]):
        with tf.get_default_graph().control_dependencies([step]):
            with tf.get_default_graph().control_dependencies([save_weights]):
                return tf.no_op()

  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")
  def variables(self):
      return super().variables() + self.optimizer.variables()



from hypergan.samplers.base_sampler import BaseSampler

from hypergan.samplers.began_sampler import BeganSampler
from hypergan.samplers.batch_sampler import BatchSampler
from hypergan.samplers.static_batch_sampler import StaticBatchSampler
from hypergan.samplers.random_walk_sampler import RandomWalkSampler
from hypergan.samplers.segment_sampler import SegmentSampler
import numpy as np
import hypergan as hg

z = None
x = None
class IdentitySampler(BaseSampler):
    def __init__(self, gan, node, samples_per_row=8, x=None, z=None):
        self.node = node
        BaseSampler.__init__(self, gan, samples_per_row)
        self.z = None
        self.x = None

    def _sample(self,i,n):
        gan = self.gan
        z_t = gan.latent.sample
        x_t = gan.inputs.x

        if self.z is None:
            self.z = []
            self.x = []
            for i in range(n):
                self.z.append(gan.session.run(z_t))
                self.x.append(gan.session.run(x_t))

        return {
                'generator': gan.session.run(self.node, {z_t: self.z[i], x_t: self.x[i]})
        }


class DebugSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)
        x_t = gan.inputs.x
        global x
        x = gan.session.run(x_t)
        self.samplers = [
          #IdentitySampler(gan, gan.inputs.x, samples_per_row),
          #IdentitySampler(gan, gan.inputs.xb, samples_per_row),
          #IdentitySampler(gan, gan.autoencoded_x, samples_per_row),
          #StaticBatchSampler(gan, samples_per_row),
          #BatchSampler(gan, samples_per_row),
          #RandomWalkSampler(gan, samples_per_row)
        ]

        #self.samplers += [IdentitySampler(gan, tf.image.resize_images(gan.inputs.x, [128,128], method=1), samples_per_row)]
        #if hasattr(gan.generator, 'pe_layers'):
        #    self.samplers += [IdentitySampler(gan, gx, samples_per_row) for gx in gan.generator.pe_layers]
        #    pe_layers = self.gan.skip_connections.get_array("progressive_enhancement")
        if hasattr(gan, 'noise_generator'):
            self.samplers += [IdentitySampler(gan, tf.concat([gan.noise_generator.sample, gan.generator.sample, gan.noise_generator.sample + gan.generator.sample], axis=0), samples_per_row)]
        #self.samplers += 
        if hasattr(gan, 'autoencoded_x'):
          self.samplers += [IdentitySampler(gan, tf.concat([gan.inputs.x,gan.autoencoded_x], axis=0), samples_per_row)]
        if gan.config.loss['class'] == BoundaryEquilibriumLoss:
          self.samplers += [BeganSampler(gan, samples_per_row)]


        if isinstance(gan.generator, SegmentGenerator):
            self.samplers += [SegmentSampler(gan)]

        if hasattr(gan, 'seq'):
            self.samplers += [IdentitySampler(gan, tf.image.resize_images(gx, [128,128], method=1), samples_per_row) for gx in gan.seq]

        for train_hook in self.gan.train_hooks():
            if isinstance(train_hook, RollingMemoryTrainHook):
                if "is_cross_replica_context" in dir(tf.distribute) and tf.distribute.is_cross_replica_context():
                    for debug in train_hook.distributed_debug():
                        self.samplers += [IdentitySampler(gan, debug, samples_per_row)]
                else:
                    self.samplers += [IdentitySampler(gan, train_hook.mx, samples_per_row)]
                    self.samplers += [IdentitySampler(gan, train_hook.mg, samples_per_row)]
            if isinstance(train_hook, RollingMemory2TrainHook):
                if "is_cross_replica_context" in dir(tf.distribute) and tf.distribute.is_cross_replica_context():
                    for debug in train_hook.distributed_debug():
                        self.samplers += [IdentitySampler(gan, debug, samples_per_row)]
                else:
                    for v in train_hook.variables():
                        self.samplers += [IdentitySampler(gan, v, samples_per_row)]

        default = gan.generator.sample#tf.zeros_like(gan.generator.layer('gend8x8'))
        def add_samples(layer):
            layer = gan.generator.layer(layer)
            if layer is None:
                layer = default

            self.samplers.append(IdentitySampler(gan, tf.image.resize_images(layer, [128,128], method=1), 1))

        #add_samples('gend8x8')
        #add_samples('gend16x16')
        #add_samples('gend32x32')
        #add_samples('gend64x64')
        #add_samples('gend128x128')
        if hasattr(gan.discriminator, 'named_layers') and "match_support_mx" in gan.discriminator.named_layers:
            self.samplers.append(IdentitySampler(gan, tf.concat([gan.inputs.x,tf.image.resize_images(gan.discriminator.named_layers['match_support_mx'], [128,128], method=1), tf.image.resize_images(gan.discriminator.named_layers['match_support_m+x'], [128,128], method=1)],axis=0),  1) )
            self.samplers.append(IdentitySampler(gan, tf.concat([gan.generator.sample, tf.image.resize_images(gan.discriminator.named_layers['match_support_mg'], [128,128], method=1), tf.image.resize_images(gan.discriminator.named_layers['match_support_m+g'], [128,128], method=1)],axis=0),  1) ) 




    def _sample(self):
        ss = []
        n=1
        for i in range(n):
            samples = [sampler._sample(i,n)['generator'] for sampler in self.samplers]
            sample_stack = np.vstack(samples)
            ss += [sample_stack]
        all_samples = np.concatenate(ss, axis=2)

        return {
            'generator':all_samples
        }




"""
The command line interface.  Trains a directory of data.
"""
from .configuration import Configuration
from .inputs import *
from .viewer import GlobalViewer
from hypergan.gan_component import ValidationException
from hypergan.gan_component import ValidationException, GANComponent
from time import sleep
import gc
import hyperchamber as hc
import hypergan as hg
import numpy as np
import os
import os
import shutil
import sys
import sys
import tempfile
import time
class CLI:
    def __init__(self, args={}, input_config=None, gan_config=None):
        self.samples = 0
        self.should_sample=False
        self.gan_config = gan_config
        self.input_config = input_config

        args = hc.Config(args)
        self.args = args

        crop =  self.args.crop

        self.config_name = self.args.config or 'default'
        self.method = args.method or 'test'
        self.total_steps = args.steps or -1
        self.sample_every = self.args.sample_every or 100

        self.sampler_name = args.sampler
        self.sampler = None

        self.loss_every = self.args.loss_every or 1

        if (self.args.save_losses):
            import matplotlib.pyplot as plt
            self.arr = []
            self.fig,self.ax = plt.subplots()
            self.temp = 0

        self.advSavePath = os.path.abspath("saves/"+self.config_name)+"/"
        if self.args.save_file:
            self.save_file = self.args.save_file + "/"
        else:
            default_save_path = os.path.abspath("saves/"+self.config_name)
            self.save_file = default_save_path + "/"
            self.create_path(self.save_file)

        title = "[hypergan] " + self.config_name
        GlobalViewer.set_options(
            enable_menu = self.args.menu,
            title = title,
            viewer_size = self.args.viewer_size,
            enabled = self.args.viewer,
            zoom = self.args.zoom)

    def sample(self, allow_save=True):
        """ Samples to a file.  Useful for visualizing the learning process.

        If allow_save is False then saves will not be created.

        Use with:

             ffmpeg -i samples/grid-%06d.png -vcodec libx264 -crf 22 -threads 0 grid1-7.mp4

        to create a video of the learning process.
        """
        sample_file="samples/%s/%06d.png" % (self.config_name, self.samples)
        self.create_path(sample_file)
        self.lazy_create()
        sample_list = self.sampler.sample(sample_file, allow_save and self.args.save_samples)
        if allow_save:
            self.samples += 1

        return sample_list

    def lazy_create(self):
        if(self.sampler == None):
            self.sampler = self.gan.sampler_for(self.sampler_name)(self.gan, samples_per_row=self.args.width)
            if(self.sampler == None):
                raise ValidationException("No sampler found by the name '"+self.sampler_name+"'")

    def step(self):
        self.gan.step()

        if(self.gan.steps % self.sample_every == 0):
            sample_list = self.sample()

    def create_path(self, filename):
        return os.makedirs(os.path.expanduser(os.path.dirname(filename)), exist_ok=True)

    def create_input(self, blank=False):
        klass = GANComponent.lookup_function(None, self.input_config['class'])
        self.input_config["blank"]=blank
        return klass(self.input_config)

    def build(self):
        return self.gan.build()

    def serve(self, gan):
        return gan_server(self.gan.session, config)

    def sample_forever(self):
        self.gan.inputs.next()
        steps = 0
        while not self.gan.destroy and (steps <= self.args.steps or self.args.steps == -1):
            self.sample()
            steps += 1

    def train(self):
        i=0
        if(self.args.ipython):
            import fcntl
            fd = sys.stdin.fileno()
            fl = fcntl.fcntl(fd, fcntl.F_GETFL)
            fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

        self.gan = hg.GAN(config=self.gan_config, inputs=self.create_input())
        self.gan.cli = self #TODO remove this link
        self.gan.inputs.next()

        if self.gan.load(self.save_file):
            print("Model loaded")
        else:
            print("Initializing new model")

        self.sample()

        while((i < self.total_steps or self.total_steps == -1) and not self.gan.destroy):
            i+=1
            self.step()
            if self.should_sample:
                self.should_sample = False
                self.sample(False)

            if (self.args.save_every != None and
                self.args.save_every != -1 and
                self.args.save_every > 0 and
                i % self.args.save_every == 0):
                print(" |= Saving network")
                self.gan.save(self.save_file)   
                self.create_path(self.advSavePath+'advSave.txt')
                if os.path.isfile(self.advSavePath+'advSave.txt'):
                    with open(self.advSavePath+'advSave.txt', 'w') as the_file:
                        the_file.write(str(self.samples)+"\n")
            if self.args.ipython:
                self.check_stdin()
        print("Done training model.  Saving")
        self.gan.save(self.save_file)
        print("============================")
        print("HyperGAN model trained")
        print("============================")

    def check_stdin(self):
        try:
            input = sys.stdin.read()
            if input[0]=="y":
                return
            from IPython import embed
            # Misc code
            embed()

        except:
            return

    def new(self):
        if self.args.toml:
            config_format = '.toml'
        else:
            config_format = '.json'
        template = self.args.directory + config_format
        print("[hypergan] Creating new configuration file '"+template+"' based off of '"+self.config_name+config_format)
        if os.path.isfile(template):
            raise ValidationException("File exists: " + template)
        source_configuration = Configuration.find(self.config_name+config_format, config_format=config_format)
        shutil.copyfile(source_configuration, template)

        return

    def run(self):
        if self.method == 'train':
            self.train()
        elif self.method == 'build':
            self.gan = hg.GAN(config=self.gan_config, inputs=self.create_input(blank=True))
            if not self.gan.load(self.save_file):
                raise ValidationException("Could not load model: "+ self.save_file)
            else:
                if os.path.isfile(self.advSavePath+'advSave.txt'):
                    with open(self.advSavePath+'advSave.txt', 'r') as the_file:
                        content = [x.strip() for x in the_file]
                        self.samples = int(content[1])
                print("Model loaded")
            self.build()
        elif self.method == 'new':
            self.new()
        elif self.method == 'sample':
            self.gan = hg.GAN(config=self.gan_config, inputs=self.create_input(blank=False))
            if not self.gan.load(self.save_file):
                print("Initializing new model")
            else:
                if os.path.isfile(self.advSavePath+'advSave.txt'):
                    with open(self.advSavePath+'advSave.txt', 'r') as the_file:
                        content = [x.strip() for x in the_file]
                        self.samples = int(content[1])
                print("Model loaded")

            self.sample_forever()



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class WeightPenaltyTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None, name="WeightPenaltyTrainHook", memory_size=2, top_k=1):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    losses = []
    weights = self.gan.weights()
    config = hc.Config(config)
    if config.only_d:
        weights = self.discriminator.weights()
    else:
        weights = self.gan.weights()
    if "l2nn" in config.constraints:
        l2nn_penalties = []
        if len(weights) > 0:
            for w in weights:
                w = tf.reshape(w, [-1, self.ops.shape(w)[-1]])
                wt = tf.transpose(w)
                wtw = tf.matmul(wt,w)
                wwt = tf.matmul(w,wt)
                def _l(m):
                    m = tf.abs(m)
                    m = tf.reduce_sum(m, axis=0,keep_dims=True)
                    m = tf.maximum(m-1, 0)
                    m = tf.reduce_max(m, axis=1,keep_dims=True)
                    return m
                l2nn_penalties.append(tf.minimum(_l(wtw), _l(wwt)))
            l2nn_penalty = self.config.l2nn_penalty * tf.add_n(l2nn_penalties)
            if self.config.constrain_every is None:
                self.add_metric('l2nn_penalty', self.gan.ops.squash(l2nn_penalty))
            losses.append(l2nn_penalty)
    if "ortho" in config.constraints:
        penalties = []
        for w in weights:
            print("PENALTY", w)
            w = tf.reshape(w, [-1, self.ops.shape(w)[-1]])
            wt = tf.transpose(w)
            wtw = tf.matmul(wt,w)
            wwt = tf.matmul(w,wt)
            mwtw = tf.matmul(w, wtw)
            mwwt = tf.matmul(wt, wwt)
            def _l(w,m):
                l = tf.reduce_mean(tf.abs(w - m))
                l = self.ops.squash(l)
                return l
            penalties.append(tf.minimum(_l(w, mwtw), _l(wt, mwwt)))
        penalty = self.config.ortho_penalty * tf.add_n(penalties)
        if self.config.constrain_every is None:
            self.add_metric('ortho_penalty', self.gan.ops.squash(penalty))
        print("PENALTY", penalty)
        penalty = tf.reshape(penalty, [1,1])
        penalty = tf.tile(penalty, [self.gan.batch_size(), 1])
        losses.append(penalty)
    if "diversity" in config.constraints:
        #Diversity regularized adversarial training
        penalties = []
        if len(weights) > 0:
            for w in weights:
                w = tf.reshape(w, [-1, self.ops.shape(w)[-1]])
                wt = tf.transpose(w)
                wt_n = tf.math.l2_normalize(wt)
                sim = tf.matmul(wt_n, tf.transpose(wt_n))
                sim = sim * (tf.ones_like(sim)-tf.eye(self.ops.shape(sim)[0]))
                mask = tf.nn.relu(tf.abs(sim)+1e-8)
                penalty = tf.reduce_sum(tf.square(sim*mask))
                penalties.append(penalty)
            full_penalty = self.config.diversity_penalty * tf.add_n(penalties)
            if self.config.constraint_every is None:
                self.add_metric('diversity_loss', self.gan.ops.squash(full_penalty))
            losses.append(full_penalty)
    print("D_L", losses)
    losses = [self.ops.squash(_l) for _l in losses]
    self.loss = tf.add_n(losses)

  def losses(self):
    return [self.loss, self.loss]

  def after_step(self, step, feed_dict):
    pass

  def before_step(self, step, feed_dict):
    pass



from hypergan.samplers.base_sampler import BaseSampler

from hypergan.samplers.began_sampler import BeganSampler
from hypergan.samplers.batch_sampler import BatchSampler
from hypergan.samplers.static_batch_sampler import StaticBatchSampler
from hypergan.samplers.random_walk_sampler import RandomWalkSampler
from hypergan.samplers.segment_sampler import SegmentSampler
import numpy as np
import hypergan as hg
from hypergan.losses.boundary_equilibrium_loss import BoundaryEquilibriumLoss
from hypergan.generators.segment_generator import SegmentGenerator

z = None
x = None
class IdentitySampler(BaseSampler):
    def __init__(self, gan, node, samples_per_row=8, x=None, z=None):
        self.node = node
        BaseSampler.__init__(self, gan, samples_per_row)
        self.z = None
        self.x = None

    def _sample(self,i,n):
        gan = self.gan
        z_t = gan.latent.sample
        x_t = gan.inputs.x

        global z
        if z is None:
            z = []
            for i in range(n):
                z.append(gan.session.run(z_t))
        if self.x is None:
            self.x = []
            for i in range(n):
                self.x.append(gan.session.run(x_t))

        return {
                'generator': gan.session.run(self.node, {z_t: z[i], x_t: self.x[i]})
        }


class ProgressiveSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)
        x_t = gan.inputs.x
        global x
        x = gan.session.run(x_t)

        self.samplers = []
        default = tf.zeros_like(gan.generator.sample)
        def add_samples(layer):
            layer = gan.generator.layer(layer)
            if layer is None:
                layer = default

            self.samplers.append(IdentitySampler(gan, tf.image.resize_images(layer, [256,256], method=1), 1))

        add_samples('g8x8')
        add_samples('g16x16')
        add_samples('g32x32')
        add_samples('g64x64')
        add_samples('g128x128')
        add_samples('g256x256')

    def _sample(self):
        ss = []
        n=4
        for i in range(n):
            samples = [sampler._sample(i,n)['generator'] for sampler in self.samplers]
            sample_stack = np.vstack(samples)
            ss += [sample_stack]
        all_samples = np.concatenate(ss, axis=2)

        return {
            'generator':all_samples
        }




"""
There are many different types of discriminators.  This is sometimes called critic.
"""
from os.path import dirname, basename, isfile
import glob
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f)]



from torch import nn
import torch

class PixelNorm(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        return input * torch.rsqrt(torch.mean(input ** 2, dim=1, keepdim=True) + 1e-8)



import torch.nn as nn
from hypergan.layer_shape import LayerShape
import hypergan as hg

class PixelShuffle(hg.Layer):
    """
        ---
        description: 'layer pixel_shuffle for configurable component'
        ---

        # pixel_shuffle layer

        Implements PixelShuffle https://pytorch.org/docs/master/generated/torch.nn.PixelShuffle.html

        ## input size

        Any 4-d tensor of the form `[B, C, H, W]`

        ## output size

        A 4d-tensor of the form `[B, C//4, H*2, W*2]`

        ## syntax

        ```json
          "pixel_shuffle"
        ```
    """

    def __init__(self, component, args, options):
        super(PixelShuffle, self).__init__(component, args, options)
        self.shuffle = nn.PixelShuffle(2)
        self.dims = list(component.current_size.dims).copy()

    def output_size(self):
        return LayerShape(self.dims[0]//4, self.dims[1]*2, self.dims[2]*2)

    def forward(self, input, context):
        return self.shuffle(input)



import numpy as np
from PIL import Image
from hypergan.viewer import GlobalViewer
from hypergan.samplers.base_sampler import BaseSampler
import time

class StyleWalkSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8, session=None):
        BaseSampler.__init__(self, gan, samples_per_row)
        self.z = None
        self.y = None
        self.x = None
        self.step = 0
        self.steps = 30
        self.target = None
        self.z_t = gan.uniform_distribution.sample
        self.z_v = gan.session.run(self.z_t)
        self.styleb_t = gan.styleb.sample

    def _sample(self):
        gan = self.gan
        inputs_t = gan.inputs.x

        if self.z is None:
            self.input = gan.session.run(gan.inputs.x)
            batch = self.input.shape[0]
            self.input = np.reshape(self.input[0], [1, self.input.shape[1], self.input.shape[2], 3])
            self.input = np.tile(self.input, [batch,1,1,1])

            self.target = gan.random_style.eval()[0]
        else:
            self.target = self.z
        self.z = gan.random_style.eval()[0]


        g=tf.get_default_graph()
        s=np.shape(gan.random_style.eval())
        bs = s[0]
        mask = np.linspace(0., 1., num=bs)
        self.z = np.tile(np.expand_dims(np.reshape(self.z, [-1]), axis=0), [bs,1])
        targ = np.tile(np.expand_dims(np.reshape(self.target, [-1]), axis=0), [bs,1])
        print("SHAPES", np.shape(mask), np.shape(self.z), np.shape(targ))
        mask = np.tile(np.expand_dims(mask, axis=1), [1, np.shape(targ)[1]])
        z_interp = np.multiply(mask,self.z) + (1-mask)*targ
        z_interp = np.reshape(z_interp, s)
        print("Z_I", np.shape(z_interp), np.shape(self.z))
        self.z = z_interp[-1]
        with g.as_default():
            tf.set_random_seed(1)
            return {
                    'generator': gan.session.run(gan.generator.sample, feed_dict={self.z_t: np.zeros_like(self.z_v), inputs_t: self.input, self.styleb_t: z_interp})
            }

    def sample(self, path, save_samples):
        gan = self.gan

        with gan.session.as_default():

            sample = self._sample()

            data = sample['generator']
            for i in range(np.shape(data)[0]):
                sample_data = data[i:i+1]
                self.plot(sample_data, path, save_samples)
                time.sleep(0.018)

            return []

    def plot(self, image, filename, save_sample):
        """ Plot an image."""
        image = np.minimum(image, 1)
        image = np.maximum(image, -1)
        image = np.squeeze(image)
        # Scale to 0..255.
        imin, imax = image.min(), image.max()
        image = (image - imin) * 255. / (imax - imin) + .5
        image = image.astype(np.uint8)
        if save_sample:
            try:
                Image.fromarray(image).save(filename)
            except Exception as e:
                print("Warning: could not sample to ", filename, ".  Please check permissions and make sure the path exists")
                print(e)
        GlobalViewer.update(self.gan, image)



import torch
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class NegativeMomentumTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None):
      super().__init__(config=config, gan=gan, trainer=trainer)
      self.d_grads = None
      self.g_grads = None

  def gradients(self, d_grads, g_grads):
      if self.d_grads is None:
          self.d_grads = [torch.zeros_like(_g) for _g in d_grads]
          self.g_grads = [torch.zeros_like(_g) for _g in g_grads]
      
      new_d_grads = [g.clone() for g in d_grads]
      new_g_grads = [g.clone() for g in g_grads]
      d_grads = [_g - self.config.gamma * _g2 for _g, _g2 in zip(d_grads, self.d_grads)]
      g_grads = [_g - self.config.gamma * _g2 for _g, _g2 in zip(g_grads, self.g_grads)]
      self.d_grads = new_d_grads
      self.g_grads = new_g_grads

      return [d_grads, g_grads]



import torch
import torch.nn as nn

class Variational(nn.Module):
    def __init__(self, channels, filter=1, stride=1, padding=0, activation=nn.LeakyReLU):
        super(Variational, self).__init__()
        self.mu_logit = nn.Conv2d(channels, channels, filter, stride, padding, padding_mode="reflect")
        self.sigma_logit = nn.Conv2d(channels, channels, filter, stride, padding, padding_mode="reflect")

    def forward(self, x):
        sigma = self.sigma_logit(x)
        mu = self.mu_logit(x)
        z = mu + torch.exp(0.5 * sigma) * torch.randn_like(sigma)
        self.sigma = sigma.view(x.shape[0], -1)
        self.mu = mu.view(x.shape[0], -1)
        return z





import torch.nn as nn

class Reshape(nn.Module):
    def __init__(self, *dims):
        self.dims = dims
        super(Reshape, self).__init__()
    def forward(self, x):
        return x.view(x.size()[0], *self.dims)



from .base_gan import BaseGAN
from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.distributions.uniform_distribution import UniformDistribution
from hypergan.gan_component import ValidationException, GANComponent
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *
from torch.autograd import Variable
from torch.autograd import grad as torch_grad
import copy
import hyperchamber as hc
import hypergan as hg
import importlib
import json
import numpy as np
import os
import sys
import time
import torch
import torch.nn as nn
import uuid

class AliGAN(BaseGAN):
    """ 
    """
    def __init__(self, *args, **kwargs):
        BaseGAN.__init__(self, *args, **kwargs)

    def create(self):
        self.classes_count = len(self.inputs.datasets)
        self.latent = self.create_component("latent")
        self.encoder = self.create_component("encoder")
        if self.config.uz:
            self.uz = self.create_component("uz")
        if self.classes_count == 2:
            self.encoder2 = self.create_component("encoder")
        self.generator = self.create_component("generator", input=self.encoder)
        self.discriminator = self.create_component("discriminator")
        self.loss = self.create_component("loss")
        self.trainer = self.create_component("trainer")

    def g_parameters(self):
        for param in self.generator.parameters():
            yield param
        for param in self.encoder.parameters():
            yield param
        if self.config.uz:
            for param in self.uz.parameters():
                yield param
        if self.classes_count == 2:
            for param in self.encoder2.parameters():
                yield param

    def d_parameters(self):
        return self.discriminator.parameters()

    def forward_pass(self):
        E = self.encoder
        G = self.generator
        D = self.discriminator

        if self.classes_count == 1:
            x0 = self.inputs.next()
            ex0 = E(x0)
            g = G(ex0)
            d_real = D(x0, context={"z": ex0})
            d_fake = D(x0, context={"z": ex0})
            self.g = g
            self.z = ex0
            self.x = x0
        elif self.classes_count == 2:
            E2 = self.encoder2
            x0 = self.inputs.next()
            x1 = self.inputs.next(1)
            if self.config.uz:
                UZ = self.uz
                euz = UZ(self.latent.sample())
                guz = G(euz)
            else:
                ex0 = E2(x0)
                gx1 = G(ex0, context=E2.context)

            ex1 = E(x1)
            if self.config.form == 1:
                d_real = D(x1, context={"z": ex1})
                d_fake = D(gx1, context={"z": ex0})
                self.z = ex1
                self.x = x1
            elif self.config.form == 3:
                d_real = D(x1, context={"z": ex1})
                d_fake = D(gx1, context={"z": E(gx1)})
                self.z = ex1
                self.x = x1
            elif self.config.form == 2:
                d_real = D(x1, context={"z": torch.abs(torch.randn_like(ex1))})
                d_fake = D(gx1, context={"z": ex0})
            else:
                d_real = D(x1, context={"z": ex0})
                d_fake = D(gx1, context={"z": ex1})
                self.z = ex0
                self.x = x1

            #self.g = G(enc_real1)
        self.d_fake = d_fake

        return d_real, d_fake

    def forward_loss(self):
        """
            Runs a forward pass through the GAN and returns (d_loss, g_loss)
        """
        d_real, d_fake = self.forward_pass()
        d_loss, g_loss = self.loss.forward(d_real, d_fake)
        mse_criterion = nn.MSELoss()
        if self.config.vae:
            logvar = self.encoder.vae.sigma
            mu = self.encoder.vae.mu
            vae = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
            #vae = (self.config.vae_lambda or 1) * vae
            self.add_metric('vae0', vae)

            if self.classes_count == 2:
                logvar = self.encoder2.vae.sigma
                mu = self.encoder2.vae.mu
                vae1 = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
                #vae1 = (self.config.vae_lambda or 1) * vae1
                vae += vae1
                self.add_metric('vae1', vae1)
 
            g_loss += vae
        if self.config.mse:
            E = self.encoder
            G = self.generator
            mse = mse_criterion(G(E(self.x)), self.x)
            self.add_metric('mse', mse)
            g_loss += mse

        return d_loss, g_loss




import torch
from torch.autograd import grad as torch_grad
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class ExtragradientTrainHook(BaseTrainHook):
    def __init__(self, gan=None, config=None):
        super().__init__(config=config, gan=gan)
        self.relu = torch.nn.ReLU()

    def gradients(self, d_grads, g_grads):
        step_size = self.config.step_size or 1e-4
        gamma = self.config.gamma
        if gamma is None:
            gamma = 1.0
        rho = self.config.rho
        if rho is None:
            rho = 1.0

        d_grads_v1 = [g.clone().detach() for g in d_grads]
        g_grads_v1 = [g.clone().detach() for g in g_grads]

        self.step(d_grads_v1, g_grads_v1, step_size)
        d_grads_v2, g_grads_v2 = self.gan.trainer.calculate_gradients()
        self.step([-g for g in d_grads_v1], [-g for g in g_grads_v1], step_size)


        if self.config.formulation == 'v2-v1':
            d_grads = [gamma*_v1-rho*(_v2-_v1)/step_size for _v1, _v2 in zip(d_grads_v1, d_grads_v2)]
            g_grads = [gamma*_v1-rho*(_v2-_v1)/step_size for _v1, _v2 in zip(g_grads_v1, g_grads_v2)]
        elif self.config.formulation == 'agree':
            d_grads = [gamma*self.agree(_v1,_v2,gamma,rho) for _v1, _v2 in zip(d_grads_v1, d_grads_v2)]
            g_grads = [gamma*self.agree(_v1,_v2,gamma,rho) for _v1, _v2 in zip(g_grads_v1, g_grads_v2)]
        else:
            d_grads = [gamma*_v1+rho*_v2 for _v1, _v2 in zip(d_grads_v1, d_grads_v2)]
            g_grads = [gamma*_v1+rho*_v2 for _v1, _v2 in zip(g_grads_v1, g_grads_v2)]

        return [d_grads, g_grads]

    def agree(self, v1, v2, gamma, rho):
        result = (gamma*v1+rho*v2) * self.relu(torch.sign(v1*v2))
        return result

    def step(self, d_grads, g_grads, step_size):
        d_params = self.gan.d_parameters()
        g_params = self.gan.g_parameters()
        for _g, _p in zip(d_grads, d_params):
            _p.data -= step_size * _g

        for _g, _p in zip(g_grads, g_params):
            _p.data -= step_size * _g




import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class QualifiedStepTrainer(BaseTrainer):
    def create(self):
        self.hist = [0 for i in range(2)]
        config = self.config
        self.global_step = tf.train.get_global_step()
        self.mix_threshold_reached = False
        decay_function = config.decay_function
        variables = self.gan.d_vars() + self.gan.g_vars()
        self.ema = [ tf.Variable(_v) for _v in variables ]
        self.store_v = [ _v.assign(_v2) for _v,_v2 in zip(self.ema, variables) ]
        self._delegate = self.gan.create_component(config.trainer, d_vars=self.d_vars, g_vars=self.g_vars, loss=self.loss)
        self._delegate.create()
        self.slot_vars_g = self._delegate.slot_vars_g
        self.slot_vars_d = self._delegate.slot_vars_g

        self.mixg = tf.Variable(1, dtype=tf.float32)
        self.mixd = tf.Variable(1, dtype=tf.float32)
        self.combine_d = [ _v.assign((1.-self.mixd) *_ema + self.mixd*_new) for _v, _ema, _new in zip(self.gan.d_vars(), self.ema, self.gan.d_vars())]
        self.combine_g = [ _v.assign((1.-self.mixg) *_ema + self.mixg*_new) for _v, _ema, _new in zip(self.gan.g_vars(), self.ema[len(self.gan.d_vars()):], self.gan.g_vars())]

        self.candidate = [ tf.Variable(_v) for _v in variables ]
        self.store_candidate = [ _v.assign(_v2) for _v,_v2 in zip(self.candidate, variables) ]
        self.reset_discriminator = [ _v.assign(_v2) for _v,_v2 in zip(self.gan.d_vars(), self.ema) ]
        self.reset_generator = [ _v.assign(_v2) for _v,_v2 in zip(self.gan.g_vars(), self.ema[len(self.gan.d_vars()):]) ]
        self.reset_candidate_discriminator = [ _v.assign(_v2) for _v,_v2 in zip(self.gan.d_vars(), self.candidate) ]
        self.reset_candidate_generator = [ _v.assign(_v2) for _v,_v2 in zip(self.gan.g_vars(), self.candidate[len(self.gan.d_vars()):]) ]
        self.candidate_loss = self.gan.loss.sample[0] - self.gan.loss.sample[1]
        if self.config.fitness == "d_fake":
            self.candidate_loss = self.gan.loss.sample[0]
        self.g_rate = 0.0
        self.d_rate = 0.0
        self.zs = [self.gan.session.run(self.gan.latent.sample) for i in range(self.config.candidate_tests)]

    def required(self):
        return "".split()

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = self.loss 

        gan.session.run(self.store_v)
        self._delegate.step(feed_dict)
        if(self.config.turn_off_step):
            if(self.config.turn_off_step < self.current_step):
                return
            else:
                print(str(-self.current_step+self.config.turn_off_step) + " qualified steps remain")

        # b = generator at new step
        # a = generator at past step
        # 1 = discriminator at past step
        # 2 = discriminator at new step
        b2 = np.mean([gan.session.run(self.candidate_loss, {gan.latent.sample: self.zs[i]}) for i in range(self.config.candidate_tests)])
        gan.session.run(self.store_candidate)
        gan.session.run(self.reset_discriminator)
        b1 = np.mean([gan.session.run(self.candidate_loss, {gan.latent.sample: self.zs[i]}) for i in range(self.config.candidate_tests)])
        gan.session.run(self.reset_generator)
        a1 = np.mean([gan.session.run(self.candidate_loss, {gan.latent.sample: self.zs[i]}) for i in range(self.config.candidate_tests)])
        gan.session.run(self.reset_candidate_discriminator)
        a2 = np.mean([gan.session.run(self.candidate_loss, {gan.latent.sample: self.zs[i]}) for i in range(self.config.candidate_tests)])
        gan.session.run(self.reset_candidate_generator)

        #    D1 D2
        # G1 a1 b1
        # G2 a2 b2
        payoff = [[a1, b1],[a2, b2]]

        d1 = a1 + a2
        d2 = b1 + b2
        g1 = a1 + b1
        g2 = a2 + b2
        if self.config.negate:
            g1 = -g1
            g2 = -g2
        mixd = d2/(d1 + d2)
        mixg = g1/(g1 + g2)
        if self.config.find_zero:
            slope = d2 - d1
            t = -d1/slope
            print("d", t)
            mixd = t
            slope = g2 - g1
            t = -g1/slope
            print("g", t)
            mixg = t
            print(t * slope + g1)
        if self.config.reverse:
            mixd = d1/(d1 + d2)
            mixg = g2/(g1 + g2)
        if self.config.loud:
            mixd = d1/(d1 + d2)
            mixg = g1/(g1 + g2)
        if self.config.zero:
            if np.abs(d2) <= np.abs(d1):
                mixd = 1.0
                self.d_rate += 1
            else:
                mixd = 0.0
            if np.abs(g2) >= np.abs(g1):
                self.g_rate += 1
                mixg = 1.0
            else:
                mixg = 0.0
        if self.config.unbounded:
            pass
        else:
            bounds = self.config.bounds or [0.0, 1.0]
            mixd = np.minimum(bounds[1], mixd)
            mixd = np.maximum(bounds[0], mixd)
            mixg = np.minimum(bounds[1], mixg)
            mixg = np.maximum(bounds[0], mixg)
        gan.session.run([self.combine_d, self.combine_g], {self.mixd: mixd, self.mixg: mixg})

        if self.current_step % 100 == 0:
            print("rates %d/100 g %d/100 d" % (self.g_rate, self.d_rate))
            self.g_rate = 0.0
            self.d_rate = 0.0



import hyperchamber as hc

from hypergan.losses.base_loss import BaseLoss

class SoftmaxLoss(BaseLoss):
    """https://arxiv.org/abs/1704.06191"""

    def _forward(self, d_real, d_fake):
        ln_zb = (((-d_real).exp().sum()+(-d_fake).exp().sum())+1e-12).log()

        d_target = 1.0 / d_real.shape[0]
        g_target = d_target / 2.0

        g_loss = g_target * (d_fake.sum() + d_real.sum()) + ln_zb
        d_loss = d_target * d_real.sum() + ln_zb

        return [d_loss, g_loss]




#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class IMLETrainHook(BaseTrainHook):
  """https://arxiv.org/pdf/1809.09087.pdf"""
  def __init__(self, gan=None, config=None, trainer=None, name="ClosestExampleTrainHook", search_size=4, memory_size=1, new_entries=-1):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    latent = gan.latent
    if self.config.use_encoder:
        latent = gan.u_to_z
    self.x_matched = [ tf.Variable(tf.zeros_like(gan.generator.sample)) for j in range(memory_size)]
    self.latent_max = tf.Variable( tf.zeros_like(latent.sample))
    self.latent = [ tf.Variable(tf.zeros_like(latent.sample)) for j in range(memory_size)]
    self.assign_x = [tf.assign(self.x_matched[j], gan.inputs.x) for j in range(memory_size)]
    self.d_lambda = config['lambda'] or 1

    self.search_size = search_size
    self.memory_size = memory_size
    self.new_entries = new_entries
    if self.new_entries == -1:
        self.new_entries = memory_size

    if self.config.use_encoder:
        encoded = self.gan.encoder.sample
        self.assign_encoded_latent = [self.latent[j].assign(encoded) for j in range(memory_size)]
    else:
        self.assign_latent_to_min = [self.latent[j].assign(self.latent_max) for j in range(memory_size)]
    self.assign_latent = self.latent_max.assign(latent.sample)

    l2_losses = tf.zeros([1])
    self.gi = []
    for j in range(memory_size):
        self.gi.append(self.gan.create_generator(self.latent[j], reuse=True))
        diff = tf.abs(self.gi[-1].sample-self.x_matched[j])
        n = tf.norm(diff, ord=2)
        l2_losses += tf.reduce_sum(diff/n)
        
    self.l2_loss_on_saved = tf.reduce_sum(self.d_lambda * l2_losses)

    l2_losses = []
    for j in range(memory_size):
        diff = tf.abs(gan.generator.sample-self.x_matched[j])
        n = tf.norm(diff, ord=2)
        l2_losses.append(tf.reduce_sum(diff/n))

    self.l2_loss_on_g = l2_losses
    self.offset = 0

    self.gan.add_metric('perceptual', self.l2_loss_on_saved)

  def losses(self):
      return [None, self.l2_loss_on_saved]

  def after_step(self, step, feed_dict):
      pass

  def before_step(self, step, feed_dict):
    if step % (self.config.step_count or 1000) != 0:
      return

    new_entries = self.new_entries
    if step == 0:
        new_entries = self.memory_size
    print("[IMLE] recalculating likelihood")
    if self.config.use_encoder:
        for j in range(new_entries):
            _j = (j+self.offset) % self.memory_size
            self.gan.session.run([self.assign_x[_j], self.assign_encoded_latent[_j]])
    else:
        for j in range(new_entries):
            _j = (j+self.offset) % self.memory_size
            self.gan.session.run(self.assign_x[_j])
            min_score = None
            for i in range(self.search_size):
                s,_ = self.gan.session.run([self.l2_loss_on_g[_j], self.assign_latent])
                if( min_score == None or min_score > s):
                    self.gan.session.run(self.assign_latent_to_min[_j])
                    min_score = s
                else:
                    pass
                    #print("Ignoring score", s)
            print("  Min ", min_score)
    self.offset += new_entries




from tensorflow.python.framework import ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import resource_variable_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.ops import variable_scope
from tensorflow.python.training import optimizer


class TPUNegativeMomentumOptimizer(optimizer.Optimizer):
    def __init__(self, learning_rate=1.0, decay=1.0, gan=None, config=None, use_locking=False, name="TPUNegativeMomentumOptimizer", optimizer=None):
        super(TPUNegativeMomentumOptimizer, self).__init__(use_locking, name)
        self._decay = decay
        self.gan = gan
        self.config = config
        self.name = name
        self.optimizer = self.gan.create_optimizer(optimizer)

    def _prepare(self):
        self.optimizer._prepare()

    def _create_slots(self, var_list):
        super(TPUNegativeMomentumOptimizer, self)._create_slots(var_list)
        self.optimizer._create_slots(var_list)
        for v in var_list :
            self._zeros_slot(v, "nm", self._name)

    def _apply_dense(self, grad, var):
      nm = self.get_slot(var, "nm")
      new_val = grad - nm
      var_update = self.optimizer._apply_dense(new_val, var)
      with tf.control_dependencies([var_update]):
          save = tf.assign(nm, ((self.config.alpha or 0.666) *grad+ (1-self.config.beta or 0.5)*nm))
          return save

    def _resource_apply_dense(self, grad, var):
      grad = tf.to_float(grad)
      nm = self.get_slot(var, "nm")
      new_val = grad - nm
      var_update = self.optimizer._resource_apply_dense(new_val, var)
      with tf.control_dependencies([var_update]):
          save = tf.assign(nm, ((self.config.alpha or 0.666) *grad+ (1-self.config.beta or 0.5)*nm))
          return save



from hypergan.samplers.base_sampler import BaseSampler
import numpy as np
import torch


class GridSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, 8)

    def _sample(self):
        gan = self.gan
        y = np.linspace(0,1, 6)

        z = np.mgrid[-0.999:0.999:0.6, -0.999:0.999:0.26].reshape(2,-1).T
        z = np.reshape(z, [32,2])
        #z = np.mgrid[-0.499:0.499:0.3, -0.499:0.499:0.13].reshape(2,-1).T
        #z = np.mgrid[-0.299:0.299:0.15, -0.299:0.299:0.075].reshape(2,-1).T
        needed = 32 / gan.batch_size()
        gs = []
        for i in range(int(needed)):
            zi = z[i*gan.batch_size():(i+1)*gan.batch_size()]
            zi = torch.from_numpy(zi).type(torch.FloatTensor).cuda()
            g = gan.generator(zi).detach().cpu().numpy()
            gs.append(g)
        g = np.reshape(gs, [4, 8, gan.channels(), gan.height(), gan.width()])
        g = np.concatenate(g, axis=0)
        g = torch.from_numpy(g).cuda()
        #x_hat = gan.session.run(gan.autoencoded_x, feed_dict={gan.inputs.x: self.x})
        #e = gan.session.run(gan.encoder.sample, feed_dict={gan.inputs.x: g})

        return [
            ('generator',g)
            ]



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12


class LineSearchTrainer(BaseTrainer):
    def _create(self):
        self._variables = []
        gan = self.gan
        config = self.config

        loss = self.gan.loss
        d_loss, g_loss = loss.sample

        self.d_log = -tf.log(tf.abs(d_loss+TINY))
        self.d_loss = d_loss
        self.g_loss = g_loss
        self.step_ops = None

        d_vars = self.gan.discriminator.ops.weights + self.gan.discriminator.ops.biases
        g_vars = self.gan.generator.ops.weights + self.gan.generator.ops.biases
        d_grads = tf.gradients(d_loss, d_vars)
        g_grads = tf.gradients(g_loss, g_vars)
        if self.config.clarification_optimizer:
            self.clarification_optimizer = self.gan.create_optimizer(self.config.alpha_optimizer)
            clarification_apply_vec = list(zip((d_grads + g_grads), (d_vars + g_vars))).copy()
            self.clarification_optimizer.apply_gradients(clarification_apply_vec)
            d_grads, g_grads = self.clarification_optimizer.grads_and_vars[0]

        g_alpha_var = tf.Variable(self.config.initial_alpha, trainable=False, name="g_alpha_dontsave")
        d_alpha_var = tf.Variable(self.config.initial_alpha, trainable=False, name="d_alpha_dontsave")
        self.reset_d_alpha = d_alpha_var.assign(self.config.initial_alpha)
        self.reset_g_alpha = g_alpha_var.assign(self.config.initial_alpha)
        self._variables += [g_alpha_var, d_alpha_var]

        g_alpha_variables = [ _x + _g * tf.abs(g_alpha_var) for _x, _g in zip(g_vars, g_grads) ]
        d_alpha_variables = [ _x + _g * tf.abs(d_alpha_var) for _x, _g in zip(d_vars, d_grads) ]

        g_alpha_biases = g_alpha_variables[len(self.gan.generator.ops.weights):]
        g_alpha_weights = g_alpha_variables[:len(self.gan.generator.ops.weights)]
        d_alpha_biases = d_alpha_variables[len(self.gan.discriminator.ops.weights):]
        d_alpha_weights = d_alpha_variables[:len(self.gan.discriminator.ops.weights)]

        self.g_alpha = self.gan.create_component(gan.config.generator, name="generator", input=self.gan.latent.sample, weights=g_alpha_weights, biases=g_alpha_biases)
        x, g = self.gan.inputs.x, self.g_alpha.sample
        self.d_alpha = self.gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([x,g],axis=0), weights=d_alpha_weights, biases=d_alpha_biases)

        self.loss = self.gan.create_component(gan.config.loss, discriminator=self.d_alpha)
        self.alpha_d_loss, self.alpha_g_loss = self.loss.sample

        self.d_optimizer = self.gan.create_optimizer(self.config.alpha_optimizer)
        self.g_optimizer = self.gan.create_optimizer(self.config.alpha_optimizer)
        self.g_train = self.g_optimizer.minimize(self.loss.sample[1], var_list=[g_alpha_var])
        self.d_train = self.d_optimizer.minimize(self.loss.sample[0], var_list=[d_alpha_var])
        self.reset_d_alpha_optimizer = tf.variables_initializer(self.d_optimizer.variables())
        self.reset_g_alpha_optimizer = tf.variables_initializer(self.g_optimizer.variables())

        self.d_alpha_var = d_alpha_var
        self.g_alpha_var = g_alpha_var

        self.g_loss = g_loss
        self.d_loss = d_loss
        self.gan.trainer = self

        d_alpha_grads = [ _g * tf.nn.relu(d_alpha_var) for _g in d_grads ]
        g_alpha_grads = [ _g * tf.nn.relu(g_alpha_var) for _g in g_grads ]

        apply_vec = list(zip((d_alpha_grads + g_alpha_grads), (d_vars + g_vars))).copy()
        self.optimizer = self.gan.create_optimizer(config.optimizer)
        self.optimize_t = self.optimizer.apply_gradients(apply_vec)
    def required(self):
        return "".split()

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = gan.loss
        metrics = gan.metrics()

        self.before_step(self.current_step, feed_dict)

        sess.run([self.reset_d_alpha, self.reset_g_alpha, self.reset_d_alpha_optimizer, self.reset_g_alpha_optimizer])

        alpha_d, alpha_g, alpha_d_loss, alpha_g_loss = sess.run([self.d_alpha_var, self.g_alpha_var, self.alpha_d_loss, self.alpha_g_loss], feed_dict)
        for i in range(self.config.nsteps):
            old_d_loss = alpha_d_loss
            alpha_d, alpha_d_loss = sess.run([self.d_alpha_var, self.alpha_d_loss], feed_dict)
            if self.config.d_threshold and alpha_d_loss < self.config.d_threshold:
                break
            if self.config.verbose or ((i+1) % 100) == 0:
                print(i, "D Alpha", alpha_d, "loss", alpha_d_loss, "diff", (old_d_loss - alpha_d_loss) )
            sess.run([self.d_train], feed_dict)

            old_g_loss = alpha_g_loss
            alpha_g, alpha_g_loss = sess.run([self.g_alpha_var, self.alpha_g_loss], feed_dict)
            if self.config.g_threshold and alpha_g_loss < self.config.g_threshold:
                break
            if self.config.verbose or ((i+1) % 100) == 0:
                print(i, "G Alpha", alpha_g, "loss", alpha_g_loss, "diff", (old_g_loss - alpha_g_loss) )
            sess.run([self.g_train], feed_dict)
 
        alpha_d, alpha_g, alpha_d_loss, alpha_g_loss = sess.run([self.d_alpha_var, self.g_alpha_var, self.alpha_d_loss, self.alpha_g_loss], feed_dict)
        #if self.config.verbose:
        print("Final Alpha %.2e %.2e loss %.2e %.2e" % (alpha_d, alpha_g, alpha_d_loss, alpha_g_loss))
        sess.run(self.optimize_t, feed_dict)
        d_loss, g_loss = loss.sample

        self.after_step(self.current_step, feed_dict)

        if self.current_step % 10 == 0:
            metric_values = self.gan.session.run(self.output_variables(metrics))
            self.print_metrics(self.current_step)

    def print_metrics(self, step):
        metrics = self.gan.metrics()
        metric_values = self.gan.session.run(self.output_variables(metrics))
        print(str(self.output_string(metrics) % tuple([step] + metric_values)))

    def variables(self):
        return super().variables() + self._variables



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class KLipschitzTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None, name="KLipschitzTrainHook", memory_size=2, top_k=1):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    if config.ragan:
        lipschitz_penalty = tf.maximum(tf.square(d_real-d_fake) - 1, 0) + tf.maximum(tf.square(d_fake-d_real) - 1, 0)
        self.add_metric('k_lipschitz_ragan', lipschitz_penalty)
    else:
        lipschitz_penalty = tf.maximum(tf.square(d_real) - 1, 0) + tf.maximum(tf.square(d_fake) - 1, 0)
        self.add_metric('k_lipschitz', ops.squash(lipschitz_penalty))
    self.d_loss = lipschitz_penalty

  def losses(self):
    return [self.d_loss, None]

  def after_step(self, step, feed_dict):
    pass

  def before_step(self, step, feed_dict):
    pass



import hyperchamber as hc
import torch

from hypergan.losses.base_loss import BaseLoss

class WassersteinLoss(BaseLoss):

    def _forward(self, d_real, d_fake):
        config = self.config

        if config.kl:
            # https://arxiv.org/abs/1910.09779
            d_fake_norm = torch.mean(d_fake.exp())+1e-8
            d_fake_ratio = (d_fake.exp()+1e-8) / d_fake_norm
            d_fake = d_fake * d_fake_ratio

        d_loss = -d_real + d_fake
        g_loss = -d_fake

        return [d_loss, g_loss]



import torch.nn as nn
import torch

class Hardsigmoid(nn.Module):
    def __init__(self):
        super().__init__()
        self.zero = torch.Tensor([0.0]).cuda()
        self.one = torch.Tensor([1.0]).cuda()

    def forward(self, input):
        return torch.min(torch.max(input, self.zero), self.one)



"""
Various layer definitions are defined here. The difference between layers and modules are that layers:
    * Have an input and output size
    * Have a standard interface
    * Use parsed arguments from the configurable component
    * Are attached to a Component
"""
from os.path import dirname, basename, isfile
import glob
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f)]

from .operation import Operation
from .add import Add
from .cat import Cat
from .mul import Mul
from .ez_norm import EzNorm
from .residual import Residual
from .channel_attention import ChannelAttention
from .pixel_shuffle import PixelShuffle
from .upsample import Upsample
from .resizable_stack import ResizableStack
from .multi_head_attention import MultiHeadAttention
from .segment_softmax import SegmentSoftmax



import importlib
import json
import numpy as np
import os
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
from hypergan.ops import TensorflowOps
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from ..base_gan import BaseGAN

from hypergan.distributions.uniform_distribution import UniformDistribution
from hypergan.trainers.experimental.consensus_trainer import ConsensusTrainer

class AliVibGAN(BaseGAN):
    """ 
    """
    def __init__(self, *args, **kwargs):
        BaseGAN.__init__(self, *args, **kwargs)

    def required(self):
        """
        `encoder`  encodes X into Z
        `discriminator`  measures X and G.
        `generator` produces samples
        """
        return "generator discriminator ".split()
    def bottleneck(self, metric, name, term1, term2):
        dvs = []
        _inputs = term1
        inputs = tf.concat(term2, axis=0)
        features = None
        bdisc = self.create_component(config[name+'1'], name=name+'1', input=inputs, features=[features])
        dvs += bdisc.variables()
        l2 = self.create_loss(config.loss, bdisc, None, None, len(_inputs))
        self.add_metric(metric+'_dl1', l2.d_loss)
        self.add_metric(metric+'_gl1', l2.g_loss)
        dl= ib_1_c * l2.d_loss
        gl=ib_1_c * l2.g_loss

        beta = config.bottleneck_beta or 1
        _features = term2
        inputs = tf.concat(_inputs, axis=0)
        features = tf.concat(_features, axis=0)
        bdisc2 = self.create_component(config[name+'2'], name=name+'2', input=inputs, features=[features])
        dvs += bdisc2.variables()
        l2 = self.create_loss(config.loss, bdisc2, None, None, len(_inputs))
        self.add_metric(metric+'_dl2',  ib_2_c * beta * l2.d_loss)
        self.add_metric(metric+'_gl2',  ib_2_c * beta * l2.g_loss)
        dl += ib_2_c * beta * l2.d_loss
        gl += ib_2_c * beta * l2.g_loss
        return gl, dl, dvs

    def create(self):
        config = self.config
        ops = self.ops
        d_losses = []
        g_losses = []
        encoder = self.create_encoder(self.inputs.x)

        with tf.device(self.device):
            x_input = tf.identity(self.inputs.x, name='input')

            if config.u_to_z:
                latent = UniformDistribution(self, config.latent)
            else:
                z_shape = self.ops.shape(encoder.sample)
                uz_shape = z_shape
                uz_shape[-1] = uz_shape[-1] // len(config.latent.projections)
                latent = UniformDistribution(self, config.latent, output_shape=uz_shape)
            self.uniform_distribution = latent 
            self.latent = latent
            direction, slider = self.create_controls(self.ops.shape(latent.sample))
            z = latent.sample + slider * direction

            u_to_z = self.create_component(config.u_to_z, name='u_to_z', input=z)
            generator = self.create_component(config.generator, input=u_to_z.sample, name='generator')
            stacked = [x_input, generator.sample]
            self.generator = generator


            self.encoder = encoder
            features = [encoder.sample, u_to_z.sample]

            reencode_u_to_z = self.create_encoder(generator.sample, reuse=True)
            reencode_u_to_z_to_g= self.create_component(config.generator, input=reencode_u_to_z.sample, name='generator', reuse=True)

            self.reencode_g = reencode_u_to_z_to_g

            x_hat = self.create_component(config.generator, input=encoder.sample, reuse=True, name='generator').sample
            reencode_x_hat_to_z = self.create_encoder(x_hat, reuse=True)
            self.uniform_sample = generator.sample

            d_vars = []
            g_vars = generator.variables() + encoder.variables()
            g_vars += u_to_z.variables()

            def ali(*stack, reuse=False):
                xs=[t for t,_ in stack]
                zs=[t for _,t in stack]
                xs=tf.concat(xs,axis=0)
                zs=tf.concat(zs,axis=0)

                discriminator = self.create_component(config.discriminator, name='discriminator', input=xs, features=[zs], reuse=reuse)
                loss = self.create_loss(config.loss, discriminator, None, None, len(stack))
                return loss,discriminator

            def d(name, stack):
                if name is None:
                    name = config
                stacked = tf.concat(stack,axis=0)
                discriminator = self.create_component(config[name], name=name, input=stacked)
                loss = self.create_loss(config.loss, discriminator, None, None, len(stack))
                return loss,discriminator

            l1, d1 = ali([self.inputs.x,encoder.sample],[generator.sample,u_to_z.sample],[reencode_u_to_z_to_g.sample, reencode_u_to_z.sample])
            l2, d2 = ali([self.inputs.x,tf.zeros_like(encoder.sample)],[generator.sample,tf.zeros_like(u_to_z.sample)],[reencode_u_to_z_to_g.sample, tf.zeros_like(reencode_u_to_z.sample)], reuse=True)
            l3, d3 = ali([tf.zeros_like(self.inputs.x),encoder.sample],[tf.zeros_like(generator.sample),u_to_z.sample],[tf.zeros_like(reencode_u_to_z_to_g.sample), reencode_u_to_z.sample], reuse=True)


            if config.alternate:
                d_losses = [beta * (l1.d_loss - l2.d_loss - l3.d_loss) + l2.d_loss + 2*l3.d_loss]
                g_losses = [beta * (l1.g_loss - l2.g_loss - l3.g_loss) + l2.g_loss + 2*l3.g_loss]

            if config.mutual_only:
                d_losses = [2*l1.d_loss - l2.d_loss - l3.d_loss]
                g_losses = [2*l1.g_loss - l2.g_loss - l3.g_loss]
            else:

                l4, d4 = d('x_discriminator', [self.inputs.x, generator.sample, reencode_u_to_z_to_g.sample])
                l5, d5 = d('z_discriminator', [encoder.sample, u_to_z.sample, reencode_u_to_z.sample])

                beta = config.beta or 0.9
                d_losses = [beta * (l1.d_loss - l2.d_loss - l3.d_loss) + l4.d_loss + 2*l5.d_loss]
                g_losses = [beta * (l1.g_loss - l2.g_loss - l3.g_loss) + l4.g_loss + 2*l5.g_loss]
            self.discriminator = d1
            self.loss = l1


            self.add_metric("ld", d_losses[0])
            self.add_metric("lg", g_losses[0])

            if config.mutual_only or config.alternate:
                for d in [d1]:
                    d_vars += d.variables()
            else:
                for d in [d1,d4,d5]:
                    d_vars += d.variables()

            self._g_vars = g_vars
            self._d_vars = d_vars

            loss = hc.Config({
                'd_fake':l1.d_fake,
                'd_real':l1.d_real,
                'sample': [tf.add_n(d_losses), tf.add_n(g_losses)]
            })
            self.loss = loss
            self.uniform_distribution = latent 
            trainer = self.create_component(config.trainer, g_vars = g_vars, d_vars = d_vars)

            self.initialize_variables()

        self.trainer = trainer
        self.generator = generator
        self.slider = slider
        self.direction = direction
        self.z = z
        self.z_hat = encoder.sample
        self.x_input = x_input
        self.autoencoded_x = x_hat
        rgb = tf.cast((self.generator.sample+1)*127.5, tf.int32)
        self.generator_int = tf.bitwise.bitwise_or(rgb, 0xFF000000, name='generator_int')
        self.random_z = tf.random_uniform(ops.shape(latent.sample), -1, 1, name='random_z')

        if hasattr(generator, 'mask_generator'):
            self.mask_generator = generator.mask_generator
            self.mask = mask
            self.autoencode_mask = generator.mask_generator.sample
            self.autoencode_mask_3_channel = generator.mask

    def get_layer(self, name):
        return self.discriminator.named_layers[name]

    def fitness_inputs(self):
        return [
                self.uniform_distribution.sample
                ]


    def create_loss(self, loss_config, discriminator, x, generator, split, reuse=False):
        loss = self.create_component(loss_config, discriminator = discriminator, x=x, split=split, reuse=reuse)
        return loss

    def create_controls(self, z_shape):
        direction = tf.random_normal(z_shape, stddev=0.3, name='direction')
        slider = tf.get_variable('slider', initializer=tf.constant_initializer(0.0), shape=[1, 1], dtype=tf.float32, trainable=False)
        return direction, slider

    def create_encoder(self, x_input, name='encoder', reuse=False):
        config = self.config
        encoder = dict(config.encoder or config.g_encoder or config.generator)
        encoder = self.create_component(encoder, name=name, input=x_input, reuse=reuse)
        return encoder

    def create_z_discriminator(self, z, z_hat):
        config = self.config
        z_discriminator = dict(config.z_discriminator or config.discriminator)
        z_discriminator['layer_filter']=None
        net = tf.concat(axis=0, values=[z, z_hat])
        encoder_discriminator = self.create_component(z_discriminator, name='z_discriminator', input=net)
        return encoder_discriminator

    def create_cycloss(self, x_input, x_hat):
        config = self.config
        ops = self.ops
        distance = config.distance or ops.lookup('l1_distance')
        pe_layers = self.gan.skip_connections.get_array("progressive_enhancement")
        cycloss_lambda = config.cycloss_lambda
        if cycloss_lambda is None:
            cycloss_lambda = 10
        
        if(len(pe_layers) > 0):
            mask = self.progressive_growing_mask(len(pe_layers)//2+1)
            cycloss = tf.reduce_mean(distance(mask*x_input,mask*x_hat))

            cycloss *= mask
        else:
            cycloss = tf.reduce_mean(distance(x_input, x_hat))

        cycloss *= cycloss_lambda
        return cycloss


    def create_z_cycloss(self, z, x_hat, encoder, generator):
        config = self.config
        ops = self.ops
        total = None
        distance = config.distance or ops.lookup('l1_distance')
        if config.z_hat_lambda:
            z_hat_cycloss_lambda = config.z_hat_cycloss_lambda
            recode_z_hat = encoder.reuse(x_hat)
            z_hat_cycloss = tf.reduce_mean(distance(z_hat,recode_z_hat))
            z_hat_cycloss *= z_hat_cycloss_lambda
        if config.z_cycloss_lambda:
            recode_z = encoder.reuse(generator.reuse(z))
            z_cycloss = tf.reduce_mean(distance(z,recode_z))
            z_cycloss_lambda = config.z_cycloss_lambda
            if z_cycloss_lambda is None:
                z_cycloss_lambda = 0
            z_cycloss *= z_cycloss_lambda

        if config.z_hat_lambda and config.z_cycloss_lambda:
            total = z_cycloss + z_hat_cycloss
        elif config.z_cycloss_lambda:
            total = z_cycloss
        elif config.z_hat_lambda:
            total = z_hat_cycloss
        return total



    def input_nodes(self):
        "used in hypergan build"
        if hasattr(self.generator, 'mask_generator'):
            extras = [self.mask_generator.sample]
        else:
            extras = []
        return extras + [
                self.x_input,
                self.slider, 
                self.direction,
                self.uniform_distribution.sample
        ]


    def output_nodes(self):
        "used in hypergan build"

    
        if hasattr(self.generator, 'mask_generator'):
            extras = [
                self.mask_generator.sample, 
                self.generator.g1x,
                self.generator.g2x
            ]
        else:
            extras = []
        return extras + [
                self.encoder.sample,
                self.generator.sample, 
                self.uniform_sample,
                self.generator_int,
                self.random_z
        ]

    def g_vars(self):
        return self._g_vars
    def d_vars(self):
        return self._d_vars



import math
import torch
import torch.nn as nn

import torch.nn.functional as F

class StyledConv2d(nn.Module):
    def __init__(
            self,
            in_channel,
            out_channel,
            kernel_size,
            style_size,
            demodulate=True,
            upsample=True,
            downsample=False,
            lr_mul=1.0,
            blur_kernel=[1, 3, 3, 1],
            ):

        super(StyledConv2d, self).__init__()
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.kernel_size = kernel_size
        fan_in = in_channel * kernel_size ** 2
        self.scale = 1 / math.sqrt(fan_in)
        self.padding = kernel_size // 2
        self.weight = nn.Parameter(
            torch.randn(out_channel, in_channel, kernel_size, kernel_size)
        )
        self.beta = nn.Linear(style_size, self.in_channel)

    def forward(self, input, style):
        batch, in_channel, height, width = input.shape
        weight = self.scale * self.weight
        #print("shape", batch, weight.shape, self.beta(style).view(batch, self.out_channel, self.kernel_size, self.kernel_size).shape)
        weight += self.beta(style)[0].view(1, self.in_channel, 1, 1)
        out = F.conv2d(input, weight, padding=self.padding)

        return out




import hypergan as hg
import hyperchamber as hc
import numpy as np

from hypergan.losses.boundary_equilibrium_loss import BoundaryEquilibriumLoss
from hypergan.losses.wasserstein_loss import WassersteinLoss
from hypergan.losses.least_squares_loss import LeastSquaresLoss
from hypergan.losses.softmax_loss import SoftmaxLoss
from hypergan.losses.standard_loss import StandardLoss
from hypergan.losses.lamb_gan_loss import LambGanLoss

from hypergan.search.random_search import RandomSearch

class AlphaGANRandomSearch(RandomSearch):
    def __init__(self, overrides):
        self.options = {
            'g_encoder': self.discriminator(),
            'z_discriminator': self.discriminator(),
            'discriminator': self.discriminator(),
            'generator': self.generator(),
            'trainer': self.trainer(),
            'loss':self.loss(),
            'encoder':self.encoder()
         }

        alpha_options = {
            'g_encoder_layers': [2,3,4,5],
            'z_discriminator_layers': [0,1,2],
            'z_discriminator_extra_layers': [0,1,2],
            'z_discriminator_extra_layers_reduction': [1,2],
            'cycloss_lambda': [0.1, 0.3, 0.2],
            'concat_linear': [64,128,256],
            'concat_linear_filters': [32,64,128,256],
            'skip_linear': [False, True],
            'd_layer_filter': [True,False],
            'g_layer_filter': [True,False],
            'encode_layer_filter': [True, False]
        }

        alpha_config = hc.Selector(alpha_options).random_config()

        self.options['g_encoder']['layers']=alpha_config.g_encoder_layers
        self.options['z_discriminator']['layers']=alpha_config.z_discriminator_layers
        self.options['z_discriminator']['extra_layers']=alpha_config.z_discriminator_extra_layers
        self.options['z_discriminator']['extra_layers_reduction']=alpha_config.z_discriminator_extra_layers_reduction
        self.options['cycloss_lambda']=alpha_config.cycloss_lambda
        self.options['generator']['concat_linear']=alpha_config.concat_linear
        self.options['generator']['concat_linear_filters']=alpha_config.concat_linear_filters
        self.options['generator']['skip_linear']=alpha_config.skip_linear
        self.options["class"]="class:hypergan.gans.alpha_gan.AlphaGAN"
        self.options['d_layer_filter']=alpha_config.d_layer_filter
        self.options['g_layer_filter']=alpha_config.g_layer_filter
        self.options = {**self.options, **overrides}



import torch.nn as nn
import torch

class Zeros(nn.Module):
    def __init__(self, shape):
        super(Zeros, self).__init__()
        self.zeros = torch.zeros(shape).cuda()

    def forward(self, net):
        return self.zeros



from hypergan.viewer import GlobalViewer
import hyperchamber as hc
import numpy as np
import inspect
from hypergan.gan_component import ValidationException, GANComponent
from torch.autograd import grad as torch_grad
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

from hypergan.viewer import GlobalViewer

class InitializeAsAutoencoder(BaseTrainHook):
    """
        G becomes an autoencoder on step zero.
    """
    def __init__(self, gan=None, config=None, trainer=None):
        super().__init__(config=config, gan=gan, trainer=trainer)

    def before_step(self, step, feed_dict, depth=0):
        if self.gan.steps != 1:
            return

        defn = self.config.encoder.copy()
        klass = GANComponent.lookup_function(None, defn['class'])
        encode = klass(self.gan, defn).cuda()
        defn = self.config.optimizer.copy()
        klass = GANComponent.lookup_function(None, defn['class'])
        del defn["class"]
        self.optimizer = klass(list(encode.parameters()) + list(self.gan.generator.parameters()), **defn)

        for i in range(self.config.steps or 1000):
            self.optimizer.zero_grad()
            inp = self.gan.inputs.next()
            e = encode(inp)
            fake = self.gan.generator(e)

            loss = ((inp - fake)**2).mean()
            loss.backward()
            for p in (list(self.gan.g_parameters())+list(encode.parameters())):
                p.requires_grad = True
            #move = torch_grad(outputs=loss, inputs=list(self.gan.g_parameters())+list(encode.parameters()), retain_graph=True, create_graph=True)
            #print(list(self.gan.g_parameters())+list(encode.parameters()))
            self.optimizer.step()
            if self.config.verbose:
                print("[autoencode]", i, "loss", loss.item())
            if self.config.info and (i % 100) == 0:
                print("[autoencode]", i, "loss", loss.item())



import hyperchamber as hc
import torch

from hypergan.losses.base_loss import BaseLoss

TINY=1e-12
class RaganLoss(BaseLoss):
    """https://arxiv.org/abs/1807.00734"""

    def required(self):
        return "".split()

    def _forward(self, d_real, d_fake):
        config = self.config
        gan = self.gan
        loss_type = self.config.type or "standard"

        if config.rgan:
            cr = d_real
            cf = d_real
        else:
            cr = torch.mean(d_real,0)
            cf = torch.mean(d_fake,0)

        if loss_type == "least_squares":
            a,b,c = (config.labels or [-1,1,1])
            d_loss = 0.5*(d_real - cf - b)**2 + 0.5*(d_fake - cr - a)**2
            g_loss = 0.5*(d_fake - cr - c)**2 + 0.5*(d_real - cf - a)**2
        elif loss_type == "hinge":
            d_loss = torch.clamp(1-(d_real - cf), min=0)+torch.clamp(1+(d_fake-cr), min=0)
            g_loss = torch.clamp(1-(d_fake - cr), min=0)+torch.clamp(1+(d_real-cf), min=0)
        elif loss_type == "wasserstein":
            d_loss = -(d_real-cf) + (d_fake-cr)
            g_loss = -(d_fake-cr)
        elif loss_type == "standard":
            d_loss = -tf.log(tf.nn.sigmoid(d_real-cf)+TINY)
            g_loss = -tf.log(tf.nn.sigmoid(d_fake-cr)+TINY)

        return [d_loss, g_loss]

    def forward_adversarial_norm(self, d_real, d_fake):
        return self.relu((d_real - d_fake.mean()).mean() - (d_fake-d_real.mean()).mean()) ** 2



import torch.nn as nn
import hypergan as hg
from hypergan.layer_shape import LayerShape

from hypergan.modules.modulated_conv2d import EqualLinear

class Upsample(hg.Layer):

    """
        ---
        description: 'layer upsample for configurable component'
        ---

        # upsample layer

        `upsample` resizes the input tensor to the specified size.

        ## Optional arguments

            * `h` - requested height. defaults to input height * 2
            * `w` - requested width. defaults to input width * 2

        ## input size

        Any 4-d tensor

        ## output size

        [B, input channels, h, w]

        ## syntax

        ```json
          "upsample"
        ```

        ## examples

        ```json
          "upsample w=96 h=96",
          "conv 4",
          "hardtanh"
        ```
    """
    def __init__(self, component, args, options):
        super(Upsample, self).__init__(component, args, options)
        w = options.w or component.current_size.width * 2
        h = options.h or component.current_size.height * 2
        self.layer = nn.Upsample((h, w), mode="bilinear")
        self.size = LayerShape(component.current_size.channels, h, w)

    def forward(self, input, context):
        return self.layer(input)

    def output_size(self):
        return self.size



import pyparsing
from pyparsing import alphas, alphanums, delimitedList, replaceWith, nestedExpr, oneOf, pyparsing_common, Dict, Forward, Keyword, Group, Suppress, Word, ZeroOrMore

class Pattern:
    def __init__(self, tokens):
        self.__dict__.update(tokens.asDict())
        args = list(tokens[1:-1])
        if args != [[]]:
            self.args = args

    def to_list(self):
        return [self.layer_name, list(self.args), self.options]

class Parser:
    def __init__(self):
        FALSE = Keyword("false")
        NULL = Keyword("null")
        TRUE = Keyword("true")
        FALSE.setParseAction(replaceWith(False))
        NULL.setParseAction(replaceWith(None))
        TRUE.setParseAction(replaceWith(True))
        pattern = Forward()
        label = Word(alphas, alphanums+"_").setResultsName("layer_name")
        configurable_param = nestedExpr(content = pattern)
        arg = (NULL ^ FALSE ^ TRUE ^ pyparsing_common.number ^ (Word(alphanums+"*_") + ~ Word("=")) ^ configurable_param)
        args = arg[...].setResultsName("args")
        args.setParseAction(self.convert_list)
        options = Dict(Group(Word(alphanums+"_") + Suppress("=") + arg))[...].setResultsName("options")
        options.setParseAction(self.convert_dict)
        pattern <<= label + args + options
        pattern.setParseAction(Pattern)
        self.pattern = pattern

    def convert_dict(self, s, l, toks):
        retv = {}
        for sublist in toks:
            retv[sublist[0]] = sublist[1]
        return retv

    def convert_list(self, s, l, toks):
        if len(toks) == 0:
            return [[]]

        return list(toks)

    def parse_string(self, string):

        parsed = self.pattern.parseString(string, parseAll=True)
        return parsed[0]



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class MultiStepTrainer(BaseTrainer):
    def __init__(self, gan, config, losses=[], var_lists=[], metrics=None):
        self.losses = losses
        self.var_lists = var_lists
        self.metrics = metrics or [None for i in self.losses]
        BaseTrainer.__init__(self, gan, config)

    def _create(self):
        gan = self.gan
        config = self.config
        losses = self.losses

        optimizers = []
        for i, _ in enumerate(losses):
            loss = losses[i][1]
            var_list = self.var_lists[i]
            is_generator = 'generator' in losses[i][0]

            if is_generator:
                optimizer = self.build_optimizer(config, 'g_', config.g_trainer, self.g_lr, var_list, loss)
            else:
                optimizer = self.build_optimizer(config, 'd_', config.d_trainer, self.d_lr, var_list, loss)
            optimizers.append(optimizer)

        self.optimizers = optimizers


        if config.d_clipped_weights:
            self.clip = [tf.assign(d,tf.clip_by_value(d, -config.d_clipped_weights, config.d_clipped_weights))  for d in d_vars]
        else:
            self.clip = []

        return None

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        losses = self.losses
        metrics = self.metrics

        for i, _ in enumerate(losses):
            loss = losses[i]
            optimizer = self.optimizers[i]
            metric = metrics[i]
            if(metric):
                metric_values = sess.run([optimizer] + self.output_variables(metric), feed_dict)[1:]

                if self.current_step % 100 == 0:
                    print("loss " + str(i) + "  "+ loss[0] + " " + self.output_string(metric) % tuple([self.current_step] + metric_values))
            else:
                _ = sess.run(optimizer, feed_dict)



import importlib
import json
import numpy as np
import os
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from ..base_gan import BaseGAN

from hypergan.distributions.uniform_distribution import UniformDistribution
from hypergan.trainers.experimental.consensus_trainer import ConsensusTrainer

class AlphaGAN(BaseGAN):
    """ 
      AlphaGAN, or α-GAN from https://arxiv.org/pdf/1706.04987.pdf
    """
    def __init__(self, *args, **kwargs):
        BaseGAN.__init__(self, *args, **kwargs)

    def required(self):
        """
        `input_encoder` is a discriminator.  It encodes X into Z
        `z_discriminator` is another discriminator.  It takes as input the output of input_encoder and z
        `discriminator` is a standard discriminator.  It measures X, reconstruction of X, and G.
        `generator` produces two samples, input_encoder output and a known random distribution.
        """
        return "generator discriminator z_discriminator ".split()

    def create(self):
        config = self.config
        ops = self.ops

        with tf.device(self.device):
            x_input = tf.identity(self.inputs.x, name='input')

            encoder = self.create_encoder(x_input)
            self.encoder = encoder
            z_shape = self.ops.shape(encoder.sample)

            uz_shape = z_shape
            uz_shape[-1] = uz_shape[-1] // len(config.encoder.projections)
            latent = UniformDistribution(self, config.encoder, output_shape=uz_shape)
            direction, slider = self.create_controls(self.ops.shape(latent.sample))
            z = latent.sample + slider * direction
            
            #projected_encoder = UniformDistribution(self, config.encoder, z=encoder.sample)

            z_discriminator = self.create_z_discriminator(latent.sample, encoder.sample)

            feature_dim = len(ops.shape(z))-1
            #stack_z = tf.concat([encoder.sample, z], feature_dim)
            #stack_encoded = tf.concat([encoder.sample, encoder.sample], feature_dim)
            stack_z = z

            generator = self.create_component(config.generator, input=stack_z)
            self.uniform_sample = generator.sample
            x_hat = generator.reuse(encoder.sample)

            if hasattr(generator, 'mask_single_channel'):
                mask = generator.mask_single_channel

            encoder_loss = self.create_loss(config.eloss or config.loss, z_discriminator, z, encoder, 2)
            if config.segments_included:
                newsample = generator.reuse(stack_z, mask=generator.mask_single_channel)
#                stacked = [x_input, generator.sample, newsample, x_hat]
                stacked = [x_input, newsample, generator.sample, x_hat, generator.g1x, generator.g2x, generator.g3x]
                #stacked = [x_input, g1x, g2x, newsample, generator.sample, x_hat]
                #stacked = [x_input, newsample, generator.sample, x_hat]
            elif config.simple_d:
                stacked = [x_input, self.uniform_sample]
            else:
                stacked = [x_input, self.uniform_sample, x_hat]

            stacked_xg = ops.concat(stacked, axis=0)
            standard_discriminator = self.create_component(config.discriminator, name='discriminator', input=stacked_xg)
            standard_loss = self.create_loss(config.loss, standard_discriminator, x_input, generator, len(stacked))
            self.loss = standard_loss

            #loss terms
            cycloss = self.create_cycloss(x_input, x_hat)
            z_cycloss = self.create_z_cycloss(latent.sample, encoder.sample, encoder, generator)

            #first_pixel = tf.slice(generator.mask_single_channel, [0,0,0,0], [-1,1,1,-1]) + 1 # we want to minimize range -1 to 1
            #cycloss += tf.reduce_sum(tf.reshape(first_pixel, [-1]), axis=0)

            if hasattr(generator, 'mask'): #TODO only segment
                cycloss_whitening_lambda = config.cycloss_whitening_lambda or 0.01
                cycloss += tf.reduce_mean(tf.reshape(0.5-tf.abs(generator.mask-0.5), [-1]), axis=0) * cycloss_whitening_lambda

            #if hasattr(generator, 'mask'): # TODO only multisegment
            #    cycloss_single_channel_lambda = config.cycloss_single_channel_lambda or 0.01
            #    m = tf.reduce_sum(generator.mask, 3)
            #    cycloss += tf.reduce_mean(tf.reshape(tf.abs(1.0-m)/ops.shape(generator.mask)[3], [-1]), axis=0) * cycloss_single_channel_lambda
            if hasattr(generator, 'mask'): # TODO only multisegment
            #    cycloss_single_channel_lambda = config.cycloss_single_channel_lambda or 0.01
                m = tf.reduce_mean(generator.mask, 1, keep_dims=True)
                m = tf.reduce_mean(m, 2, keep_dims=True)
                c = 0.1
                cycloss += (c - tf.minimum(tf.reduce_min(m, 3, keep_dims=True), c))
            #    cycloss += tf.reduce_mean(tf.reshape(tf.abs(1.0-m)/ops.shape(generator.mask)[3], [-1]), axis=0) * cycloss_single_channel_lambda

            trainer = self.create_trainer(cycloss, z_cycloss, encoder, generator, encoder_loss, standard_loss, standard_discriminator, z_discriminator)
            self.initialize_variables()

        self.trainer = trainer
        self.generator = generator
        self.uniform_distribution = uniform_encoder
        self.slider = slider
        self.direction = direction
        self.z = z
        self.z_hat = encoder.sample
        self.x_input = x_input
        self.autoencoded_x = x_hat
        rgb = tf.cast((self.generator.sample+1)*127.5, tf.int32)
        self.generator_int = tf.bitwise.bitwise_or(rgb, 0xFF000000, name='generator_int')
        self.random_z = tf.random_uniform(ops.shape(UniformDistribution.sample), -1, 1, name='random_z')

        if hasattr(generator, 'mask_generator'):
            self.mask_generator = generator.mask_generator
            self.mask = mask
            self.autoencode_mask = generator.mask_generator.sample
            self.autoencode_mask_3_channel = generator.mask

    def create_loss(self, loss_config, discriminator, x, generator, split):
        loss = self.create_component(loss_config, discriminator = discriminator, x=x, generator=generator.sample, split=split)
        return loss

    def create_controls(self, z_shape):
        direction = tf.random_normal(z_shape, stddev=0.3, name='direction')
        slider = tf.get_variable('slider', initializer=tf.constant_initializer(0.0), shape=[1, 1], dtype=tf.float32, trainable=False)
        return direction, slider

    def create_encoder(self, x_input, name='input_encoder'):
        config = self.config
        input_encoder = dict(config.input_encoder or config.g_encoder or config.discriminator)
        encoder = self.create_component(input_encoder, name=name, input=x_input)
        return encoder

    def create_z_discriminator(self, z, z_hat):
        config = self.config
        z_discriminator = dict(config.z_discriminator or config.discriminator)
        z_discriminator['layer_filter']=None
        net = tf.concat(axis=0, values=[z, z_hat])
        encoder_discriminator = self.create_component(z_discriminator, name='z_discriminator', input=net)
        return encoder_discriminator

    def create_cycloss(self, x_input, x_hat):
        config = self.config
        ops = self.ops
        distance = config.distance or ops.lookup('l1_distance')
        pe_layers = self.gan.skip_connections.get_array("progressive_enhancement")
        cycloss_lambda = config.cycloss_lambda
        if cycloss_lambda is None:
            cycloss_lambda = 10
        
        if(len(pe_layers) > 0):
            mask = self.progressive_growing_mask(len(pe_layers)//2+1)
            cycloss = tf.reduce_mean(distance(mask*x_input,mask*x_hat))

            cycloss *= mask
        else:
            cycloss = tf.reduce_mean(distance(x_input, x_hat))

        cycloss *= cycloss_lambda
        return cycloss


    def create_z_cycloss(self, z, x_hat, encoder, generator):
        config = self.config
        ops = self.ops
        total = None
        distance = config.distance or ops.lookup('l1_distance')
        if config.z_hat_lambda:
            z_hat_cycloss_lambda = config.z_hat_cycloss_lambda
            recode_z_hat = encoder.reuse(x_hat)
            z_hat_cycloss = tf.reduce_mean(distance(z_hat,recode_z_hat))
            z_hat_cycloss *= z_hat_cycloss_lambda
        if config.z_cycloss_lambda:
            recode_z = encoder.reuse(generator.reuse(z))
            z_cycloss = tf.reduce_mean(distance(z,recode_z))
            z_cycloss_lambda = config.z_cycloss_lambda
            if z_cycloss_lambda is None:
                z_cycloss_lambda = 0
            z_cycloss *= z_cycloss_lambda

        if config.z_hat_lambda and config.z_cycloss_lambda:
            total = z_cycloss + z_hat_cycloss
        elif config.z_cycloss_lambda:
            total = z_cycloss
        elif config.z_hat_lambda:
            total = z_hat_cycloss
        return total



    def create_trainer(self, cycloss, z_cycloss, encoder, generator, encoder_loss, standard_loss, standard_discriminator, encoder_discriminator):
        if z_cycloss is not None:
            loss1=('generator encoder', z_cycloss + cycloss + encoder_loss.g_loss)
            loss2=('generator image', z_cycloss + cycloss + standard_loss.g_loss)
            loss3=('discriminator image', standard_loss.d_loss)
            loss4=('discriminator encoder', encoder_loss.d_loss)
        else:
            loss1=('generator encoder', cycloss + encoder_loss.g_loss)
            loss2=('generator image',cycloss + standard_loss.g_loss)
            loss3=('discriminator image', standard_loss.d_loss)
            loss4=('discriminator encoder', encoder_loss.d_loss)

        var_lists = []
        var_lists.append(encoder.variables())
        var_lists.append(generator.variables())
        var_lists.append(standard_discriminator.variables())
        var_lists.append(encoder_discriminator.variables())

        metrics = []
        metrics.append(None)
        metrics.append(None)
        metrics.append(standard_loss.metrics)
        metrics.append(encoder_loss.metrics)

        trainer = self.create_component(self.config.trainer)
        return trainer

    def input_nodes(self):
        "used in hypergan build"
        if hasattr(self.generator, 'mask_generator'):
            extras = [self.mask_generator.sample]
        else:
            extras = []
        return extras + [
                self.x_input,
                self.slider, 
                self.direction,
                self.uniform_distribution.sample
        ]


    def output_nodes(self):
        "used in hypergan build"

    
        if hasattr(self.generator, 'mask_generator'):
            extras = [
                self.mask_generator.sample, 
                self.generator.g1x,
                self.generator.g2x
            ]
        else:
            extras = []
        return extras + [
                self.encoder.sample,
                self.generator.sample, 
                self.uniform_sample,
                self.generator_int,
                self.random_z
        ]



from hypergan.samplers.base_sampler import BaseSampler

from hypergan.samplers.began_sampler import BeganSampler
from hypergan.samplers.batch_sampler import BatchSampler
from hypergan.samplers.static_batch_sampler import StaticBatchSampler
from hypergan.samplers.random_walk_sampler import RandomWalkSampler
from hypergan.samplers.segment_sampler import SegmentSampler
import numpy as np
import hypergan as hg

class SortedSampler(BaseSampler):
    def __init__(self, gan):
        BaseSampler.__init__(self, gan)
        self.xs = None
        self.samples = 10
        self.display_count = 5

    def sample(self, path, sample_to_file):
        gan = self.gan

        sess = gan.session
        config = gan.config
        if self.xs is None:
            self.xs = [sess.run(gan.fitness_inputs()) for i in range(self.samples)]

        current_g = sess.run(gan.trainer.all_g_vars)
        
        stacks = []
        def _samples():
            cs = []
            for i in range(self.samples):
                ts = gan.fitness_inputs()
                vs = self.xs[i]
                feed_dict = {}
                for t,v in zip(ts, vs):
                    feed_dict[t]=v
                cs.append(sess.run(gan.generator.sample,feed_dict))
                #cs.append(sess.run(gan.autoencoded_x,feed_dict))
            return cs
        
        gs = _samples()
        priority = []
        for i, sample in zip(range(self.samples), gs):
            priority.append(gan.session.run(gan.loss.d_fake, {gan.generator.sample: sample}))

        sorted_sgs = [[p, v] for p,v in zip(priority, gs)]
        sorted_sgs.sort(key=lambda x: -x[0])
        sorted_sgs = [s[1] for s in sorted_sgs]

        top = sorted_sgs[:self.display_count]
        end = sorted_sgs[(len(gs)-self.display_count):]

        print("TOP", np.shape(top), "end", np.shape(end))
        stacks.append(np.vstack(top))
        stacks.append(np.vstack(end))
        images = np.vstack([np.hstack(s) for s in stacks])

        self.plot(images, path, sample_to_file)
        return [{'image': path, 'label': 'tiled x sample'}]



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class MultiMarginalTrainer(BaseTrainer):
    def _create(self):
        gan = self.gan
        config = self.config

        loss = gan.loss
        c_loss, d_loss, g_loss = loss.sample_cdg

        self.d_log = -tf.log(tf.abs(d_loss+TINY))

        g_optimizer = config.g_optimizer or config.optimizer
        d_optimizer = config.d_optimizer or config.optimizer
        c_optimizer = config.c_optimizer or config.optimizer
        d_optimizer["loss"] = d_loss
        g_optimizer["loss"] = g_loss
        c_optimizer["loss"] = c_loss
        g_optimizer = self.gan.create_optimizer(g_optimizer)
        d_optimizer = self.gan.create_optimizer(d_optimizer)
        c_optimizer = self.gan.create_optimizer(c_optimizer)

        d_grads = tf.gradients(d_loss, gan.trainable_d_vars())
        g_grads = tf.gradients(g_loss, gan.trainable_g_vars())
        c_grads = tf.gradients(c_loss, gan.trainable_c_vars())
        #d_grads = [gan.distribution_strategy.reduce( tf.distribute.ReduceOp.SUM, grad, axis=None) for grad in d_grads]
        #g_grads = [gan.distribution_strategy.reduce( tf.distribute.ReduceOp.SUM, grad, axis=None) for grad in g_grads]
        apply_vec_g = list(zip((g_grads), (gan.trainable_g_vars()))).copy()
        apply_vec_d = list(zip((d_grads), (gan.trainable_d_vars()))).copy()
        apply_vec_c = list(zip((c_grads), (gan.trainable_c_vars()))).copy()
        self.g_loss = g_loss
        self.d_loss = d_loss
        self.gan.trainer = self
        g_optimizer_t = g_optimizer.apply_gradients(apply_vec_g)
        d_optimizer_t = d_optimizer.apply_gradients(apply_vec_d)
        c_optimizer_t = c_optimizer.apply_gradients(apply_vec_c)

        self.d_optimizer = d_optimizer
        self.d_optimizer_t = d_optimizer_t
        self.g_optimizer = g_optimizer
        self.g_optimizer_t = g_optimizer_t
        self.c_optimizer = c_optimizer
        self.c_optimizer_t = c_optimizer_t

        return g_optimizer, d_optimizer

    def variables(self):
        return self.ops.variables() + self.d_optimizer.variables() + self.g_optimizer.variables() + self.c_optimizer.variables()

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = gan.loss
        metrics = gan.metrics()

        self.before_step(self.current_step, feed_dict)
        for i in range(config.d_update_steps or 1):
            sess.run([self.d_optimizer_t], feed_dict)
            sess.run([self.c_optimizer_t], feed_dict)

        metric_values = sess.run([self.g_optimizer_t] + self.output_variables(metrics), feed_dict)[1:]
        self.after_step(self.current_step, feed_dict)

        if self.current_step % 100 == 0:
            print(str(self.output_string(metrics) % tuple([self.current_step] + metric_values)))

    def print_metrics(self, step):
        metrics = self.gan.metrics()
        metric_values = self.gan.session.run(self.output_variables(metrics))
        print(str(self.output_string(metrics) % tuple([step] + metric_values)))

    def distributed_step(self):
        config = self.config
        ops = []
        for i in range(config.d_update_steps or 1):
            ops += [self.d_optimizer_t]
        with tf.control_dependencies(ops):
            return self.g_optimizer_t



import numpy as np
import hyperchamber as hc
import inspect
from hypergan.trainers.base_trainer import BaseTrainer

class ProportionalControlTrainer(BaseTrainer):

    def _create(self):
        d_loss = gan.graph.d_loss
        g_loss = gan.graph.g_loss
        g_lr = np.float32(config.g_learn_rate)
        d_lr = np.float32(config.d_learn_rate)

        gan.graph.d_vars = d_vars
        g_defk = {k[2:]: v for k, v in config.items() if k[2:] in inspect.getargspec(config.g_trainer).args and k.startswith("d_")}
        d_defk = {k[2:]: v for k, v in config.items() if k[2:] in inspect.getargspec(config.d_trainer).args and k.startswith("g_")}
        g_optimizer = config.g_trainer(g_lr, **g_defk)
        d_optimizer = config.d_trainer(d_lr, **d_defk)
        if(config.clipped_gradients):
            g_optimizer = capped_optimizer(g_optimizer, config.clipped_gradients, g_loss, g_vars)
            d_optimizer = capped_optimizer(d_optimizer, config.clipped_gradients, d_loss, d_vars)
        else:
            g_optimizer = g_optimizer.minimize(g_loss, var_list=g_vars)
            d_optimizer = d_optimizer.minimize(d_loss, var_list=d_vars)

        gan.graph.clip = [tf.assign(d,tf.clip_by_value(d, -config.d_clipped_weights, config.d_clipped_weights))  for d in d_vars]

        return g_optimizer, d_optimizer


    def _step(gan, feed_dict):
        sess = gan.sess
        config = gan.config
        x_t = gan.graph.x
        g_t = gan.graph.g
        d_log_t = gan.graph.d_log
        g_loss = gan.graph.g_loss
        d_loss = gan.graph.d_loss
        g_optimizer = gan.graph.g_optimizer
        d_optimizer = gan.graph.d_optimizer
        d_class_loss = gan.graph.d_class_loss
        d_vars = gan.graph.d_vars

        # in WGAN paper, values are clipped.  This might not work, and is slow.
        if(config.d_clipped_weights):
            sess.run(gan.graph.clip)

        _, d_cost, d_log = sess.run([d_optimizer, d_loss, d_log_t], feed_dict)

        _, g_cost, g_k, measure= sess.run([g_optimizer, g_loss, gan.graph.update_k, gan.graph.measure], feed_dict)
        if self.step % 100 == 0:
            print("%2d: g cost %.2f d_loss %.2f k %.2f m %.2f gamma %.2f" % (self.step, g_cost , d_cost, g_k, measure, gan.graph.gamma))

        self.step+=1

        return d_cost, g_cost



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class MaxGpTrainHook(BaseTrainHook):
  "https://arxiv.org/pdf/1902.05687v2.pdf C.2"
  def __init__(self, gan=None, config=None, trainer=None, name="MaxGpTrainHook", memory_size=2, top_k=1):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    gan_inputs = self.gan.inputs.x
    if hasattr(self.gan.inputs, 'frames'):
        gan_inputs = tf.concat(self.gan.inputs.frames[1:], axis=3)
        latent_sample = self.gan.c0
    else:
        latent_sample = self.gan.latent.sample
    self.s_max = [ tf.Variable( tf.zeros_like(gan_inputs)) for i in range(memory_size)]
    self.d_lambda = config['lambda'] or 1

    self.assign_s_max_new_entries = [ tf.assign(self.s_max[i], self.gan.sample_mixture()) for i in range(memory_size) ]
    self.memory_size = memory_size
    self.top_k = top_k

    self.current = tf.Variable(tf.zeros_like(gan_inputs))
    d = self.gan.create_component(self.gan.config.discriminator, name='discriminator', input=self.current, features=[tf.zeros_like(latent_sample)], reuse=True)
    self.assign_current = [ self.current.assign(self.s_max[i]) for i in range(memory_size) ]
    gd = tf.gradients(d.sample, gan.d_vars())
    r = tf.add_n([tf.square(tf.norm(_gd, ord=2)) for _gd in gd])
    self.d_loss = self.d_lambda * tf.reduce_mean(r)
    self.gan.add_metric('gpsn', self.d_loss)
    if self.config.from_source:
        self.d_loss = tf.add_n([tf.reduce_sum(tf.square(_gd)) for _gd in gd])

  def losses(self):
    return [self.d_loss, None]

  def after_step(self, step, feed_dict):
    pass

  def before_step(self, step, feed_dict):
    # get (memory_size - topk) x_hats
    for i,s in enumerate(self.assign_s_max_new_entries[self.top_k:]):
        self.gan.session.run(s)
    # sort memory
    scores = []
    for i in range(self.memory_size):
        self.gan.session.run(self.assign_current[i])
        s = self.gan.session.run(self.d_loss)
        scores.append(s)
    sort = zip(scores, self.s_max, self.assign_s_max_new_entries)
    sort2 = sorted(sort, key=itemgetter(0), reverse=True)
    new_s_max = [s_max for _, s_max,_ in sort2]
    new_assign = [a for _, _,a in sort2]
    self.s_max = new_s_max
    self.assign_s_max_new_entries = new_assign
    # get max
    if self.config.all:
        winner = sum(scores)
    else:
        winner = scores[np.argmax(scores)]

    # truncate memory to top_k
    feed_dict[self.d_loss] = winner




import hyperchamber as hc
import torch

from hypergan.losses.base_loss import BaseLoss
import torch.nn.functional as F

class StandardLoss(BaseLoss):
    def __init__(self, gan, config):
        super(StandardLoss, self).__init__(gan, config)
        self.relu = torch.nn.ReLU()
        self.two = torch.Tensor([2.0]).cuda()
        self.eps = torch.Tensor([1e-12]).cuda()

    def _forward(self, d_real, d_fake):
        criterion = torch.nn.BCEWithLogitsLoss()
        g_loss = criterion(d_fake, torch.ones_like(d_fake))
        d_loss = criterion(d_real, torch.ones_like(d_real)) + criterion(d_fake, torch.zeros_like(d_fake))

        return [d_loss, g_loss]




#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hypergan.gan_component import GANComponent
import hyperchamber as hc
import inspect

class BaseTrainHook(GANComponent):
  def __init__(self, gan=None, config=None, trainer=None):
    super().__init__(gan, config)

  def create(self):
    pass

  def losses(self):
    return [None, None]

  def before_step(self, step, feed_dict):
    pass

  def after_step(self, step, feed_dict):
    pass

  def after_create(self):
    pass

  def update_op(self):
    return None

  def gradients(self, d_grads, g_grads):
    return [d_grads, g_grads]

  def forward(self, d_loss, g_loss):
    return [None, None]



import torch.nn as nn
import hyperchamber as hc
import hypergan as hg
from hypergan.layer_shape import LayerShape

class ResizableStack(hg.Layer):
    """
        ---
        description: 'layer resizable_stack for configurable component'
        ---

        # resizable_stack layer

        `resizable_stack` allows for variable size outputs on the generator. A conv stack is repeated until the output size is reached.

        If you specify "segment_softmax" this repeats the pattern:
          upsample
          normalize(expects style vector named 'w')
          conv ...
          activation(before last layer)

        and ends in:
          segment_softmax output_channels

        ## arguments
            * layer type. Defaults to "segment_softmax"
        ## optional arguments
            * segment_channels - The number of channels before segment_softmax. Defaults to 5
            * max_channels - The most channels for any conv. Default 256
            * style - the style vector to use. Default "w"
        ## input size

        Any 4-d tensor

        ## output size

        [B, output channels, output height, output width]

        ## syntax

        ```json
          "resizable_stack segment_softmax"
        ```

        ## examples

        ```json
          "identity name=w",
          "linear 8*8*256",
          "relu",
          "resizable_stack segment_softmax",
          "hardtanh"
        ```
    """
    def __init__(self, component, args, options):
        super(ResizableStack, self).__init__(component, args, options)
        self.size = LayerShape(component.gan.channels(), component.gan.height(), component.gan.width())
        self.max_channels = options.max_channels or 256
        self.segment_channels = options.segment_channels or 5
        self.style = options.style or "w"

        layers = []

        sizes = self.sizes(component.current_size.height, component.current_size.width, component.gan.height(), component.gan.width(), self.segment_channels * 2 * component.gan.channels())
        print("SIZES", sizes)
        for i, size in enumerate(sizes[1:]):
            c = min(size.channels, self.max_channels)
            upsample = hg.layers.Upsample(component, [], hc.Config({"w": size.width, "h": size.height}))
            component.current_size = upsample.output_size() #TODO abstract input_size
            _, add = component.parse_layer("add self (ez_norm initializer=(xavier_normal) style=" + self.style + ")")
            _, conv = component.parse_layer("conv2d " + str(size.channels) + " padding=0 initializer=(xavier_normal)")
            layers += [upsample, add, conv]
            if i < len(sizes) - 2:
                layers += [nn.ReLU()]

        layers += [hg.layers.SegmentSoftmax(component, [component.gan.channels()], {})]
        self.layers = nn.ModuleList(layers)

    def sizes(self, initial_height, initial_width, target_height, target_width, final_channels):
        channels = []
        hs = []
        ws = []
        sizes = []

        w = initial_width
        h = initial_height
        i = 0

        channels.append(final_channels)
        while w < target_width or h < target_height:
            if i > 0:
                h-=2 #padding
                w-=2 #padding
            h*=2 #upscale
            w*=2 #upscale
            channels.append(final_channels * 2**i)
            hs.append(min(h, target_height+2))
            ws.append(min(w, target_width+2))
            i+=1

        w = initial_width
        h = initial_height

        channels.reverse()
        for c,h,w in zip(channels, hs, ws):
            sizes.append(LayerShape(c, h, w))
        return sizes

    def forward(self, input, context):
        for module in self.layers:
            if isinstance(module, hg.Layer):
                input = module(input, context)
            else:
                input = module(input)
        return input

    def output_size(self):
        return self.size




import torch.nn as nn

class Layer(nn.Module):
    def __init__(self, component, args, options):
        super(Layer, self).__init__()
        self.args = args
        self.options = options

    def forward(self, input, context):
        pass

    def output_size(self):
        pass



import hyperchamber as hc
import numpy as np
import torch
from .base_distribution import BaseDistribution

from ..gan_component import ValidationException
from torch.distributions import uniform

TINY=1e-12

class UniformDistribution(BaseDistribution):
    def __init__(self, gan, config):
        BaseDistribution.__init__(self, gan, config)
        self.current_channels = config["z"]
        self.current_input_size = config["z"]
        batch_size = gan.batch_size()
        self.shape = [batch_size, self.current_input_size]
        self.const_two = torch.Tensor([2]).cuda()
        self.const_one = torch.Tensor([1]).cuda()
        if self.config.projections is not None:
            self.current_input_size *= len(self.config.projections)
            self.current_channels *= len(self.config.projections)
        self.next()

    def create(self):
        pass

    def required(self):
        return "".split()

    def validate(self):
        errors = BaseDistribution.validate(self)
        #if(self.config.z is not None and int(self.config.z) % 2 != 0):
        #    errors.append("z must be a multiple of 2 (was %2d)" % self.config.z)
        return errors

    def lookup(self, projection):
        if callable(projection):
            return projection
        if projection == 'identity':
            return identity
        if projection == 'sphere':
            return sphere
        if projection == 'gaussian':
            return gaussian
        if projection == 'periodic':
            return periodic
        return self.lookup_function(projection)

    def sample(self):
        self.z = torch.rand(self.shape, device="cuda:0") * self.const_two - self.const_one
        if self.config.projections is None:
            return self.z
        projections = []
        for projection in self.config.projections:
            projections.append(self.lookup(projection)(self.config, self.gan, self.z))
        ps = []
        for p in projections:
            ps.append(self.z)
        return torch.cat(ps, -1)

    def next(self):
        self.instance = self.sample()
        return self.instance

def identity(config, gan, net):
    return net

def round(config, gan, net):
    net = torch.round(net)
    return net

def binary(config, gan, net):
    net = torch.gt(net, 0)
    net = torch.type(net, torch.Float)
    return net

def zero(config, gan, net):
    return torch.zeros_like(net)



import torch.nn as nn

class Residual(nn.Module):
    def __init__(self, channels, filter=3, stride=1, padding=1, activation=nn.ReLU):
        super(Residual, self).__init__()
        self.conv = nn.Conv2d(channels, channels, filter, stride, padding)
        self.activation = activation()
    def forward(self, x):
        return x + self.activation(self.conv(x))



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect

class GradientMagnitudeOptimizer(optimizer.Optimizer):
  """Projects layer gradients to a norm then multiplies by a constant"""
  def __init__(self, learning_rate=0.001, decay=0.9, gan=None, config=None, use_locking=False, name="EmaOptimizer", optimizer=None):
    super().__init__(use_locking, name)
    self._decay = decay
    self.gan = gan
    self.config = config
    self.name = name
    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    var_list = [v for _, v in grads_and_vars]
    grad_list = [g for g, _ in grads_and_vars]


    self._prepare()
    def project_gradient_layer(gs):
        if self.config.norm == 'softmax':
            return tf.nn.softmax(gs)
        elif self.config.norm == 'euclidean':
            return gs / (tf.sqrt(tf.reduce_sum(tf.square(gs)))+1e-8)
        elif self.config.norm == 'inf':
            return gs / (tf.norm(gs, ord=np.inf)+1e-8)
        elif self.config.norm == 'max':
            return gs / (tf.reduce_max(tf.abs(gs))+1e-8)
        elif self.config.norm == False:
            return gs
        else:
            return gs / (tf.norm(gs, ord=self.config.norm)+1e-8)

    lam = []
    for g, v in grads_and_vars:
        _lam = self.gan.configurable_param(self.config["lambda"])
        opts = self.gan.layer_options(v)
        if opts is not None:
            print("OPTS", opts)
            if "gradient_magnitude_lambda" in opts:
                _lam *= float(opts["gradient_magnitude_lambda"])
        lam.append(_lam)

    print("Lambdas = ", lam)

    def number_weights(v):
        count = np.prod(self.gan.ops.shape(v))
        return count
    if self.config.per_weight:
        newlam = []
        for _lam, v in zip(lam,grad_list):
            newlam.append(_lam * number_weights(v))
        lam = newlam
    newgrads = [_lam * project_gradient_layer(g) for _lam, g in zip(lam, grad_list)]
    newgrads_and_vars = list(zip(newgrads, var_list)).copy()
    op = self.optimizer.apply_gradients(newgrads_and_vars, global_step=global_step, name=name)
    with tf.get_default_graph().control_dependencies([op]):
        return tf.no_op()

  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")

  def variables(self):
      return super().variables() + self.optimizer.variables()



from subprocess import call

import glob

filenames = glob.glob("*.json")

for f in filenames:
    cmd = "jq '.loss.gradient_locally_stable=0.01' "+f+" | awk 'BEGIN{RS=\"\";getline<\"-\";print>ARGV[1]}' "+f
    print(cmd)
    
    result = call(cmd, shell=True)



